from . import adelie_core as core
from . import glm
from . import logger
from . import matrix
from .state import (
    gaussian_cov as state_gaussian_cov,
    gaussian_naive as state_gaussian_naive,
    glm_naive as state_glm_naive,
    multigaussian_naive as state_multigaussian_naive,
    multiglm_naive as state_multiglm_naive,
) 
from typing import (
    Callable,
    Union,
)
import numpy as np


def _solve(
    state, 
    progress_bar: bool =False,
    exit_cond: Callable =None,
):
    # mapping of each state type to the corresponding solver
    f_dict = {
        # cov methods
        core.state.StateGaussianPinCov64: core.solver.solve_gaussian_pin_cov_64,
        core.state.StateGaussianPinCov32: core.solver.solve_gaussian_pin_cov_32,
        core.state.StateGaussianCov64: core.solver.solve_gaussian_cov_64,
        core.state.StateGaussianCov32: core.solver.solve_gaussian_cov_32,
        # naive methods
        core.state.StateGaussianPinNaive64: core.solver.solve_gaussian_pin_naive_64,
        core.state.StateGaussianPinNaive32: core.solver.solve_gaussian_pin_naive_32,
        core.state.StateGaussianNaive64: core.solver.solve_gaussian_naive_64,
        core.state.StateGaussianNaive32: core.solver.solve_gaussian_naive_32,
        core.state.StateMultiGaussianNaive64: core.solver.solve_multigaussian_naive_64,
        core.state.StateMultiGaussianNaive32: core.solver.solve_multigaussian_naive_32,
        core.state.StateGlmNaive64: core.solver.solve_glm_naive_64,
        core.state.StateGlmNaive32: core.solver.solve_glm_naive_32,
        core.state.StateMultiGlmNaive64: core.solver.solve_multiglm_naive_64,
        core.state.StateMultiGlmNaive32: core.solver.solve_multiglm_naive_32,
    }

    is_gaussian_pin = (
        isinstance(state, core.state.StateGaussianPinCov64) or
        isinstance(state, core.state.StateGaussianPinCov32) or
        isinstance(state, core.state.StateGaussianPinNaive64) or
        isinstance(state, core.state.StateGaussianPinNaive32)
    )
    is_gaussian = (
        isinstance(state, core.state.StateGaussianCov64) or
        isinstance(state, core.state.StateGaussianCov32) or
        isinstance(state, core.state.StateGaussianNaive64) or
        isinstance(state, core.state.StateGaussianNaive32) or
        isinstance(state, core.state.StateMultiGaussianNaive64) or
        isinstance(state, core.state.StateMultiGaussianNaive32)
    )
    is_glm = (
        isinstance(state, core.state.StateGlmNaive64) or
        isinstance(state, core.state.StateGlmNaive32) or
        isinstance(state, core.state.StateMultiGlmNaive64) or
        isinstance(state, core.state.StateMultiGlmNaive32)
    )

    # solve group elastic net
    f = f_dict[state._core_type]
    if is_gaussian_pin:
        out = f(state)
    elif is_gaussian:
        out = f(state, progress_bar, exit_cond)
    elif is_glm:
        out = f(state, state._glm, progress_bar, exit_cond)
    else:
        raise RuntimeError("Unexpected state type.")

    # raise any errors
    if out["error"] != "":
        if out["error"].startswith("adelie_core solver: "):
            logger.logger.warning(RuntimeError(out["error"]))
        else:
            logger.logger.error(RuntimeError(out["error"]))

    # return a subsetted Python result object
    core_state = out["state"]
    state = type(state).create_from_core(state, core_state)

    # add extra total time information
    state.total_time = out["total_time"]

    return state


def gaussian_cov(
    A: np.ndarray,
    v: np.ndarray,
    *,
    constraints: list =None,
    groups: np.ndarray =None,
    alpha: float =1,
    penalty: np.ndarray =None,
    lmda_path: np.ndarray =None,
    max_iters: int =int(1e5),
    tol: float =1e-7,
    rdev_tol: float =1e-3,
    newton_tol: float =1e-12,
    newton_max_iters: int =1000,
    n_threads: int =1,
    early_exit: bool =True,
    screen_rule: str ="pivot",
    min_ratio: float =1e-2,
    lmda_path_size: int =100,
    max_screen_size: int =None,
    max_active_size: int =None,
    pivot_subset_ratio: float =0.1,
    pivot_subset_min: int =1,
    pivot_slack_ratio: float =1.25,
    check_state: bool =False,
    progress_bar: bool =True,
    warm_start =None,
    exit_cond: Callable =None,
):
    """Solves Gaussian group elastic net via covariance method.

    The Gaussian elastic net problem via covariance method minimizes the following:

    .. math::
        \\begin{align*}
            \\mathrm{minimize}_{\\beta} \\quad&
            \\frac{1}{2} \\beta^\\top A \\beta - v^\\top \\beta
            + \\lambda \\sum\\limits_{g=1}^G \\omega_g \\left(
                \\alpha \\|\\beta_g\\|_2 + \\frac{1-\\alpha}{2} \\|\\beta_g\\|_2^2
            \\right)
        \\end{align*}

    where 
    :math:`\\beta` is the coefficient vector,
    :math:`A` is any positive semi-definite matrix,
    :math:`v` is any vector,
    :math:`\\lambda \\geq 0` is the regularization parameter,
    :math:`G` is the number of groups,
    :math:`\\omega \\geq 0` is the penalty factor,
    :math:`\\alpha \\in [0,1]` is the elastic net parameter,
    and :math:`\\beta_g` are the coefficients for the :math:`g` th group.

    Parameters
    ----------
    A : (p, p) Union[adelie.matrix.MatrixCovBase64, adelie.matrix.MatrixCovBase32]
        Positive semi-definite matrix.
        It is typically one of the matrices defined in ``adelie.matrix`` submodule.
    v : (p,) np.ndarray
        Linear term.
    constraints : (G,) list, optional
        List of constraints for each group.
        ``constraints[i]`` is the constraint object corresponding to group ``i``.
        If ``constraints[i]`` is ``None``, then the ``i`` th group is unconstrained.
        If ``None``, every group is unconstrained.
        Default is ``None``.
    groups : (G,) np.ndarray, optional
        List of starting indices to each group where `G` is the number of groups.
        ``groups[i]`` is the starting index of the ``i`` th group. 
        Default is ``None``, in which case it is set to ``np.arange(p)``.
    alpha : float, optional
        Elastic net parameter.
        It must be in the range :math:`[0,1]`.
        Default is ``1``.
    penalty : (G,) np.ndarray, optional
        Penalty factor for each group in the same order as ``groups``.
        It must be a non-negative vector.
        Default is ``None``, in which case, it is set to ``np.sqrt(group_sizes)``.
    lmda_path : (L,) np.ndarray, optional
        The regularization path to solve for.
        The full path is not considered if ``early_exit`` is ``True``.
        It is recommended that the path is sorted in decreasing order.
        If ``None``, the path will be generated.
        Default is ``None``.
    max_iters : int, optional
        Maximum number of coordinate descents.
        Default is ``int(1e5)``.
    tol : float, optional
        Coordinate descent convergence tolerance.
        Default is ``1e-7``.
    rdev_tol : float, optional
        Relative percent deviance explained tolerance.
        Default is ``1e-3``.
    newton_tol : float, optional
        Convergence tolerance for the BCD update.
        Default is ``1e-12``.
    newton_max_iters : int, optional
        Maximum number of iterations for the BCD update.
        Default is ``1000``.
    n_threads : int, optional
        Number of threads.
        Default is ``1``.
    early_exit : bool, optional
        ``True`` if the function should early exit based on training deviance explained.
        Default is ``True``.
    min_ratio : float, optional
        The ratio between the largest and smallest :math:`\\lambda` in the regularization sequence
        if it is to be generated.
        Default is ``1e-2``.
    lmda_path_size : int, optional
        Number of regularizations in the path if it is to be generated.
        Default is ``100``.
    screen_rule : str, optional
        The type of screening rule to use. It must be one of the following options:

            - ``"strong"``: adds groups whose active scores are above the strong threshold.
            - ``"pivot"``: adds groups whose active scores are above the pivot cutoff with slack.

        Default is ``"pivot"``.
    max_screen_size : int, optional
        Maximum number of screen groups allowed.
        The function will return a valid state and guarantees to have screen set size
        less than or equal to ``max_screen_size``.
        If ``None``, it will be set to the total number of groups.
        Default is ``None``.
    max_active_size : int, optional
        Maximum number of active groups allowed.
        The function will return a valid state and guarantees to have active set size
        less than or equal to ``max_active_size``.
        If ``None``, it will be set to the total number of groups.
        Default is ``None``.
    pivot_subset_ratio : float, optional
        If screening takes place, then the ``(1 + pivot_subset_ratio) * s``
        largest active scores are used to determine the pivot point
        where ``s`` is the current screen set size.
        It is only used if ``screen_rule="pivot"``.
        Default is ``0.1``.
    pivot_subset_min : int, optional
        If screening takes place, then at least ``pivot_subset_min``
        number of active scores are used to determine the pivot point.
        It is only used if ``screen_rule="pivot"``.
        Default is ``1``.
    pivot_slack_ratio : float, optional
        If screening takes place, then ``pivot_slack_ratio``
        number of groups with next smallest (new) active scores 
        below the pivot point are also added to the screen set as slack.
        It is only used if ``screen_rule="pivot"``.
        Default is ``1.25``.
    check_state : bool, optional 
        ``True`` is state should be checked for inconsistencies before calling solver.
        Default is ``False``.

        .. warning::
            The check may take a long time if the inputs are big!

    progress_bar : bool, optional
        ``True`` to enable progress bar.
        Default is ``True``.
    warm_start : optional
        If no warm-start is provided, the initial solution is set to 0
        and other invariance quantities are set accordingly.
        Otherwise, the warm-start is used to extract all necessary state variables.
        If warm-start is used, the user *must* still provide consistent inputs,
        that is, warm-start will not overwrite most arguments passed into this function.
        However, changing configuration settings such as tolerance levels is well-defined.
        Default is ``None``.

        .. note::
            The primary use-case is when a user already called the function with ``warm_start=False``
            but would like to continue fitting down a longer path of regularizations.
            This way, the user does not have to restart the fit at the beginning,
            but can simply continue from the last returned state.

        .. warning::
            We have only tested warm-starts in the setting described in the note above,
            that is, when ``lmda_path`` and possibly static configurations have changed.
            Use with caution in other settings!

    exit_cond : Callable, optional
        If not ``None``, it must be a callable object that takes in a single argument.
        The argument is the current state object of the same type as the return value.
        During the optimization, 
        after obtaining the solution at each regularization value,
        ``exit_cond(state)`` is evaluated as an opportunity
        for the user to early exit the program based on their own rule.
        Default is ``None``.

        .. note::
            The algorithm early exits if ``exit_cond(state)``
            evaluates to ``True`` *or* the built-in early exit
            function evaluates to ``True`` (if ``early_exit`` is ``True``).

    Returns
    -------
    state
        The resulting state after running the solver.

    See Also
    --------
    adelie.adelie_core.state.StateGaussianCov64
    """
    if isinstance(A, np.ndarray):
        A = matrix.dense(A, method="cov", n_threads=n_threads)

    assert (
        isinstance(A, matrix.MatrixCovBase64) or
        isinstance(A, matrix.MatrixCovBase32)
    )

    dtype = (
        np.float64
        if isinstance(A, matrix.MatrixCovBase64) else
        np.float32
    )
    
    p = A.cols()

    if not (lmda_path is None):
        # MUST evaluate the flip to be able to pass into C++ backend.
        lmda_path = np.array(np.flip(np.sort(lmda_path)))

    if groups is None:
        groups = np.arange(p, dtype=int)
    group_sizes = np.concatenate([groups, [p]], dtype=int)
    group_sizes = group_sizes[1:] - group_sizes[:-1]

    G = len(groups)

    if penalty is None:
        penalty = np.sqrt(group_sizes)

    if warm_start is None:
        lmda = np.inf
        lmda_max = None
        screen_set = np.arange(G)[(penalty <= 0) | (alpha <= 0)]
        screen_beta = np.zeros(np.sum(group_sizes[screen_set]), dtype=dtype)
        screen_is_active = np.ones(screen_set.shape[0], dtype=bool)
        screen_dual_size = 0
        if not (constraints is None):
            screen_dual_size = np.sum([
                0 if constraints[k] is None else constraints[k].dual_size
                for k in screen_set
            ], dtype=int)
        screen_dual = np.zeros(screen_dual_size, dtype=dtype)
        active_set_size = screen_set.shape[0]
        active_set = np.empty(G, dtype=int)
        active_set[:active_set_size] = np.arange(active_set_size)
        rsq = 0

        subset = np.array([
            np.arange(groups[ss], groups[ss] + group_sizes[ss]) 
            for ss in screen_set
        ])
        order = np.argsort(subset)
        indices = subset[order]
        values = screen_beta[order]

        grad = np.empty(p, dtype=dtype)
        A.mul(indices, values, grad)
        grad = v - grad

    else:
        lmda = warm_start.lmda
        lmda_max = warm_start.lmda_max
        screen_set = warm_start.screen_set
        screen_beta = warm_start.screen_beta
        screen_is_active = warm_start.screen_is_active
        screen_dual = warm_start.screen_dual
        active_set_size = warm_start.active_set_size
        active_set = warm_start.active_set
        rsq = warm_start.rsq
        grad = warm_start.grad

    state = state_gaussian_cov(
        A=A,
        v=v,
        constraints=constraints,
        groups=groups,
        group_sizes=group_sizes,
        alpha=alpha,
        penalty=penalty,
        screen_set=screen_set,
        screen_beta=screen_beta,
        screen_is_active=screen_is_active,
        screen_dual=screen_dual,
        active_set_size=active_set_size,
        active_set=active_set,
        rsq=rsq,
        lmda=lmda,
        grad=grad,
        lmda_path=lmda_path,
        lmda_max=lmda_max,
        max_iters=max_iters,
        tol=tol,
        rdev_tol=rdev_tol,
        newton_tol=newton_tol,
        newton_max_iters=newton_max_iters,
        n_threads=n_threads,
        early_exit=early_exit,
        screen_rule=screen_rule,
        min_ratio=min_ratio,
        lmda_path_size=lmda_path_size,
        max_screen_size=max_screen_size,
        max_active_size=max_active_size,
        pivot_subset_ratio=pivot_subset_ratio,
        pivot_subset_min=pivot_subset_min,
        pivot_slack_ratio=pivot_slack_ratio,
    )

    if check_state:
        state.check(method="assert")

    return _solve(
        state=state, 
        progress_bar=progress_bar,
        exit_cond=exit_cond,
    )


def grpnet(
    X: np.ndarray,
    glm: Union[glm.GlmBase32, glm.GlmBase64],
    *,
    constraints: list =None,
    groups: np.ndarray =None,
    alpha: float =1,
    penalty: np.ndarray =None,
    offsets: np.ndarray =None,
    lmda_path: np.ndarray =None,
    irls_max_iters: int =int(1e4),
    irls_tol: float =1e-7,
    max_iters: int =int(1e5),
    tol: float =1e-7,
    adev_tol: float =0.9,
    ddev_tol: float =0,
    newton_tol: float =1e-12,
    newton_max_iters: int =1000,
    n_threads: int =1,
    early_exit: bool =True,
    intercept: bool =True,
    screen_rule: str ="pivot",
    min_ratio: float =1e-2,
    lmda_path_size: int =100,
    max_screen_size: int =None,
    max_active_size: int =None,
    pivot_subset_ratio: float =0.1,
    pivot_subset_min: int =1,
    pivot_slack_ratio: float =1.25,
    check_state: bool =False,
    progress_bar: bool =True,
    warm_start =None,
    exit_cond: Callable =None,
):
    """Solves group elastic net via naive method.

    The group elastic net problem minimizes the following:

    .. math::
        \\begin{align*}
            \\mathrm{minimize}_{\\beta, \\beta_0} \\quad&
            \\ell(\\eta)
            + \\lambda \\sum\\limits_{g=1}^G \\omega_g \\left(
                \\alpha \\|\\beta_g\\|_2 + \\frac{1-\\alpha}{2} \\|\\beta_g\\|_2^2
            \\right)
            \\\\
            \\text{subject to} \\quad&
            \\eta = X\\beta + \\beta_0 \\mathbf{1} + \\eta^0
        \\end{align*}

    where 
    :math:`\\beta_0` is the intercept,
    :math:`\\beta` is the coefficient vector,
    :math:`X` is the feature matrix,
    :math:`\\eta^0` is a fixed offset vector,
    :math:`\\lambda \\geq 0` is the regularization parameter,
    :math:`G` is the number of groups,
    :math:`\\omega \\geq 0` is the penalty factor,
    :math:`\\alpha \\in [0,1]` is the elastic net parameter,
    :math:`\\beta_g` are the coefficients for the :math:`g` th group,
    and :math:`\\ell(\\cdot)` is the loss function defined by a GLM.

    For multi-response problems (i.e. when :math:`y` is 2-dimensional)
    such as in multigaussian or multinomial,
    the group elastic net problem minimizes the following:

    .. math::
        \\begin{align*}
            \\mathrm{minimize}_{\\beta, \\beta_0} \\quad&
            \\ell(\\eta)
            + \\lambda \\sum\\limits_{g=1}^G \\omega_g \\left(
                \\alpha \\|\\beta_g\\|_2 + \\frac{1-\\alpha}{2} \\|\\beta_g\\|_2^2
            \\right)
            \\\\
            \\text{subject to} \\quad&
            \\mathrm{vec}(\\eta^\\top) 
            = 
            (X\\otimes I_K) \\beta + (\\mathbf{1}\\otimes I_K) \\beta_0 
            + 
            \\mathrm{vec}(\\eta^{0\\top})
        \\end{align*}

    where :math:`\\mathrm{vec}(\\cdot)` is the operator that flattens the input as column-major.
    Note that if ``intercept`` is ``True``, then an intercept for each class is provided
    as additional unpenalized features in the data matrix and the global intercept is turned off.

    Parameters
    ----------
    X : (n, p) matrix-like
        Feature matrix.
        It is typically one of the matrices defined in ``adelie.matrix`` submodule or ``np.ndarray``.
    glm : Union[adelie.glm.GlmBase32, adelie.glm.GlmBase64, adelie.glm.GlmMultiBase32, adelie.glm.GlmMultiBase64]
        GLM object.
        It is typically one of the GLM classes defined in ``adelie.glm`` submodule.
    constraints : (G,) list, optional
        List of constraints for each group.
        ``constraints[i]`` is the constraint object corresponding to group ``i``.
        If ``constraints[i]`` is ``None``, then the ``i`` th group is unconstrained.
        If ``None``, every group is unconstrained.
        Default is ``None``.
    groups : (G,) np.ndarray, optional
        List of starting indices to each group where `G` is the number of groups.
        ``groups[i]`` is the starting index of the ``i`` th group. 
        If ``glm`` is multi-response type, then we only allow two types of groupings:

            - ``"grouped"``: coefficients for each predictor is grouped across the classes.
            - ``"ungrouped"``: every coefficient is its own group.

        Default is ``None``, in which case it is set to
        ``np.arange(p)`` if ``y`` is single-response
        and ``"grouped"`` if multi-response.
    alpha : float, optional
        Elastic net parameter.
        It must be in the range :math:`[0,1]`.
        Default is ``1``.
    penalty : (G,) np.ndarray, optional
        Penalty factor for each group in the same order as ``groups``.
        It must be a non-negative vector.
        Default is ``None``, in which case, it is set to ``np.sqrt(group_sizes)``.
    offsets : (n,) or (n, K) np.ndarray, optional
        Observation offsets :math:`\\eta^0`.
        Default is ``None``, in which case, it is set to 
        ``np.zeros(n)`` if ``y`` is single-response
        and ``np.zeros((n, K))`` if multi-response.
    lmda_path : (L,) np.ndarray, optional
        The regularization path to solve for.
        The full path is not considered if ``early_exit`` is ``True``.
        It is recommended that the path is sorted in decreasing order.
        If ``None``, the path will be generated.
        Default is ``None``.
    irls_max_iters : int, optional
        Maximum number of IRLS iterations.
        This parameter is only used if ``glm`` is not of gaussian type.
        Default is ``int(1e4)``.
    irls_tol : float, optional
        IRLS convergence tolerance.
        This parameter is only used if ``glm`` is not of gaussian type.
        Default is ``1e-7``.
    max_iters : int, optional
        Maximum number of coordinate descents.
        Default is ``int(1e5)``.
    tol : float, optional
        Coordinate descent convergence tolerance.
        Default is ``1e-7``.
    adev_tol : float, optional
        Percent deviance explained tolerance.
        Default is ``0.9``.
    ddev_tol : float, optional
        Difference in percent deviance explained tolerance.
        Default is ``0``.
    newton_tol : float, optional
        Convergence tolerance for the BCD update.
        Default is ``1e-12``.
    newton_max_iters : int, optional
        Maximum number of iterations for the BCD update.
        Default is ``1000``.
    n_threads : int, optional
        Number of threads.
        Default is ``1``.
    early_exit : bool, optional
        ``True`` if the function should early exit based on training deviance explained.
        Default is ``True``.
    min_ratio : float, optional
        The ratio between the largest and smallest :math:`\\lambda` in the regularization sequence
        if it is to be generated.
        Default is ``1e-2``.
    lmda_path_size : int, optional
        Number of regularizations in the path if it is to be generated.
        Default is ``100``.
    intercept : bool, optional 
        ``True`` if the function should fit with intercept.
        If ``y`` is multi-response, then an intercept for each class is added
        and the global intercept is turned off.
        Default is ``True``.
    screen_rule : str, optional
        The type of screening rule to use. It must be one of the following options:

            - ``"strong"``: adds groups whose active scores are above the strong threshold.
            - ``"pivot"``: adds groups whose active scores are above the pivot cutoff with slack.

        Default is ``"pivot"``.
    max_screen_size : int, optional
        Maximum number of screen groups allowed.
        The function will return a valid state and guarantees to have screen set size
        less than or equal to ``max_screen_size``.
        If ``None``, it will be set to the total number of groups.
        Default is ``None``.
    max_active_size : int, optional
        Maximum number of active groups allowed.
        The function will return a valid state and guarantees to have active set size
        less than or equal to ``max_active_size``.
        If ``None``, it will be set to the total number of groups.
        Default is ``None``.
    pivot_subset_ratio : float, optional
        If screening takes place, then the ``(1 + pivot_subset_ratio) * s``
        largest active scores are used to determine the pivot point
        where ``s`` is the current screen set size.
        It is only used if ``screen_rule="pivot"``.
        Default is ``0.1``.
    pivot_subset_min : int, optional
        If screening takes place, then at least ``pivot_subset_min``
        number of active scores are used to determine the pivot point.
        It is only used if ``screen_rule="pivot"``.
        Default is ``1``.
    pivot_slack_ratio : float, optional
        If screening takes place, then ``pivot_slack_ratio``
        number of groups with next smallest (new) active scores 
        below the pivot point are also added to the screen set as slack.
        It is only used if ``screen_rule="pivot"``.
        Default is ``1.25``.
    check_state : bool, optional 
        ``True`` is state should be checked for inconsistencies before calling solver.
        Default is ``False``.

        .. warning::
            The check may take a long time if the inputs are big!

    progress_bar : bool, optional
        ``True`` to enable progress bar.
        Default is ``True``.
    warm_start : optional
        If no warm-start is provided, the initial solution is set to 0
        and other invariance quantities are set accordingly.
        Otherwise, the warm-start is used to extract all necessary state variables.
        If warm-start is used, the user *must* still provide consistent inputs,
        that is, warm-start will not overwrite most arguments passed into this function.
        However, changing configuration settings such as tolerance levels is well-defined.
        Default is ``None``.

        .. note::
            The primary use-case is when a user already called the function with ``warm_start=False``
            but would like to continue fitting down a longer path of regularizations.
            This way, the user does not have to restart the fit at the beginning,
            but can simply continue from the last returned state.

        .. warning::
            We have only tested warm-starts in the setting described in the note above,
            that is, when ``lmda_path`` and possibly static configurations have changed.
            Use with caution in other settings!

    exit_cond : Callable, optional
        If not ``None``, it must be a callable object that takes in a single argument.
        The argument is the current state object of the same type as the return value.
        During the optimization, 
        after obtaining the solution at each regularization value,
        ``exit_cond(state)`` is evaluated as an opportunity
        for the user to early exit the program based on their own rule.
        Default is ``None``.

        .. note::
            The algorithm early exits if ``exit_cond(state)``
            evaluates to ``True`` *or* the built-in early exit
            function evaluates to ``True`` (if ``early_exit`` is ``True``).

    Returns
    -------
    state
        The resulting state after running the solver.

    See Also
    --------
    adelie.adelie_core.state.StateGaussianNaive64
    adelie.adelie_core.state.StateGlmNaive64
    adelie.adelie_core.state.StateMultiGaussianNaive64
    adelie.adelie_core.state.StateMultiGlmNaive64
    """
    X_raw = X

    if isinstance(X, np.ndarray):
        X = matrix.dense(X, method="naive", n_threads=n_threads)

    assert (
        isinstance(X, matrix.MatrixNaiveBase64) or
        isinstance(X, matrix.MatrixNaiveBase32)
    )

    dtype = (
        np.float64
        if isinstance(X, matrix.MatrixNaiveBase64) else
        np.float32
    )

    n, p = X.rows(), X.cols()

    # compute common quantities
    if not (offsets is None): 
        if offsets.shape != glm.y.shape:
            raise RuntimeError("offsets must be same shape as y if not None.")
        offsets = np.array(offsets, order="C", copy=False, dtype=dtype)
    else:
        offsets = np.zeros(glm.y.shape, dtype=dtype)

    if not (lmda_path is None):
        # MUST evaluate the flip to be able to pass into C++ backend.
        lmda_path = np.array(np.flip(np.sort(lmda_path)), dtype=dtype)

    solver_args = {
        "X": X_raw,
        "constraints": constraints,
        "alpha": alpha,
        "offsets": offsets,
        "lmda_path": lmda_path,
        "max_iters": max_iters,
        "tol": tol,
        "adev_tol": adev_tol,
        "ddev_tol": ddev_tol,
        "newton_tol": newton_tol,
        "newton_max_iters": newton_max_iters,
        "n_threads": n_threads,
        "early_exit": early_exit,
        "intercept": intercept,
        "screen_rule": screen_rule,
        "min_ratio": min_ratio,
        "lmda_path_size": lmda_path_size,
        "max_screen_size": max_screen_size,
        "max_active_size": max_active_size,
        "pivot_subset_ratio": pivot_subset_ratio,
        "pivot_subset_min": pivot_subset_min,
        "pivot_slack_ratio": pivot_slack_ratio,
    }

    # do special routine for optimized gaussian
    is_gaussian_opt = (
        (glm.name in ["gaussian", "multigaussian"]) and
        glm.opt
    )

    # add a few more configs in GLM case
    if not is_gaussian_opt:
        solver_args["glm"] = glm
        solver_args["irls_max_iters"] = irls_max_iters
        solver_args["irls_tol"] = irls_tol
    else:
        solver_args["y"] = glm.y
        solver_args["weights"] = glm.weights

    # multi-response GLMs
    if glm.is_multi:
        K = glm.y.shape[-1]

        if groups is None:
            groups = "grouped"
        if groups == "grouped":
            groups = K * np.arange(p, dtype=int)
        elif groups == "ungrouped":
            groups = np.arange(K * p, dtype=int)
        else:
            raise RuntimeError(
                "groups must be one of \"grouped\" or \"ungrouped\" for multi-response."
            )
        if intercept:
            groups = np.concatenate([np.arange(K), K + groups], dtype=int)
        group_sizes = np.concatenate([groups, [(p+intercept)*K]], dtype=int)
        group_sizes = group_sizes[1:] - group_sizes[:-1]

        if penalty is None:
            penalty = np.sqrt(group_sizes).astype(dtype)
            if intercept:
                penalty[:K] = 0
        else:
            if intercept:
                penalty = np.concatenate([np.zeros(K), penalty], dtype=dtype)

        if warm_start is None:
            lmda = np.inf
            lmda_max = None
            screen_set = np.arange(groups.shape[0])[(penalty <= 0) | (alpha <= 0)]
            screen_beta = np.zeros(np.sum(group_sizes[screen_set]), dtype=dtype)
            screen_is_active = np.ones(screen_set.shape[0], dtype=bool)
            screen_dual_size = 0
            if not (constraints is None):
                screen_dual_size = np.sum([
                    0 if constraints[k] is None else constraints[k].dual_size
                    for k in screen_set
                ], dtype=int)
            screen_dual = np.zeros(screen_dual_size, dtype=dtype)
            active_set_size = screen_set.shape[0]
            active_set = np.empty(groups.shape[0], dtype=int)
            active_set[:active_set_size] = np.arange(active_set_size)
        else:
            lmda = warm_start.lmda
            lmda_max = warm_start.lmda_max
            screen_set = warm_start.screen_set
            screen_beta = warm_start.screen_beta
            screen_is_active = warm_start.screen_is_active
            screen_dual = warm_start.screen_dual
            active_set_size = warm_start.active_set_size
            active_set = warm_start.active_set

        solver_args["groups"] = groups
        solver_args["group_sizes"] = group_sizes
        solver_args["penalty"] = penalty
        solver_args["lmda"] = lmda
        solver_args["lmda_max"] = lmda_max
        solver_args["screen_set"] = screen_set
        solver_args["screen_beta"] = screen_beta
        solver_args["screen_is_active"] = screen_is_active
        solver_args["screen_dual"] = screen_dual
        solver_args["active_set_size"] = active_set_size
        solver_args["active_set"] = active_set

        # represent the augmented X matrix as used in single-response reformatted problem.
        X_aug = matrix.kronecker_eye(X_raw, K, n_threads=n_threads)
        if intercept:
            X_aug = matrix.concatenate(
                [
                    matrix.kronecker_eye(
                        np.ones((n, 1), dtype=dtype), 
                        K, 
                        n_threads=n_threads,
                    ),
                    X_aug,
                ], 
                axis=1, 
                n_threads=n_threads,
            )

        # special gaussian case
        if is_gaussian_opt:
            y = glm.y
            weights = glm.weights
            weights_mscaled = weights / K
            if warm_start is None:
                ones = np.ones(n, dtype=dtype)
                X_means = np.empty(p, dtype=dtype)
                X.mul(ones, weights_mscaled, X_means)
                X_means = np.repeat(X_means, K)
                if intercept:
                    X_means = np.concatenate([
                        np.full(K, 1/K),
                        X_means,
                    ], dtype=dtype)
                y_off = y - offsets
                # variance of y that gaussian solver expects
                y_var = np.sum(weights_mscaled[:, None] * y_off ** 2)
                # variance for the null model with multi-intercept
                # R^2 can be initialized to MSE under intercept-model minus y_var.
                # This is a negative quantity in general, but will be corrected to 0
                # when the model fits the unpenalized (including intercept) term.
                # Then, supplying y_var as the normalization will result in R^2 
                # relative to the intercept-model.
                if intercept:
                    y_off_c = y_off - (y_off.T @ weights)[None] # NOT a typo: weights
                    yc_var = np.sum(weights_mscaled[:, None] * y_off_c ** 2)
                    rsq = yc_var - y_var
                    y_var = yc_var
                else:
                    rsq = 0
                resid = y_off.ravel()
                resid_sum = np.sum(weights_mscaled[:, None] * y_off)
                grad = np.empty(X_aug.cols(), dtype=dtype)
                weights_mscaled = np.repeat(weights_mscaled, K)
                X_aug.mul(resid, weights_mscaled, grad)
            else:
                X_means = warm_start.X_means
                y_var = warm_start.y_var
                rsq = warm_start.rsq
                resid = warm_start.resid
                resid_sum = warm_start.resid_sum
                grad = warm_start.grad

            solver_args["X_means"] = X_means
            solver_args["y_var"] = y_var
            solver_args["rsq"] = rsq
            solver_args["resid"] = resid
            solver_args["resid_sum"] = resid_sum
            solver_args["grad"] = grad

            state = state_multigaussian_naive(**solver_args)
        
        # GLM case
        else:
            if warm_start is None:
                ones = np.ones(offsets.size, dtype=dtype)
                eta = offsets
                resid = np.empty(eta.shape, dtype=dtype)
                glm.gradient(eta, resid)
                resid = resid.ravel()
                grad = np.empty(X_aug.cols(), dtype=dtype)
                X_aug.mul(resid, ones, grad)
                loss_null = None
                loss_full = glm.loss_full()
                eta = eta.ravel()
            else:
                eta = warm_start.eta
                resid = warm_start.resid
                grad = warm_start.grad
                loss_null = warm_start.loss_null
                loss_full = warm_start.loss_full

            solver_args["grad"] = grad
            solver_args["eta"] = eta
            solver_args["resid"] = resid
            solver_args["loss_null"] = loss_null
            solver_args["loss_full"] = loss_full

            state = state_multiglm_naive(**solver_args)

    # single-response GLMs
    else:
        if groups is None:
            groups = np.arange(p, dtype=int)
        group_sizes = np.concatenate([groups, [p]], dtype=int)
        group_sizes = group_sizes[1:] - group_sizes[:-1]

        G = len(groups)

        if penalty is None:
            penalty = np.sqrt(group_sizes).astype(dtype)

        if warm_start is None:
            lmda = np.inf
            lmda_max = None
            screen_set = np.arange(G)[(penalty <= 0) | (alpha <= 0)]
            screen_beta = np.zeros(np.sum(group_sizes[screen_set]), dtype=dtype)
            screen_is_active = np.ones(screen_set.shape[0], dtype=bool)
            screen_dual_size = 0
            if not (constraints is None):
                screen_dual_size = np.sum([
                    0 if constraints[k] is None else constraints[k].dual_size
                    for k in screen_set
                ], dtype=int)
            screen_dual = np.zeros(screen_dual_size, dtype=dtype)
            active_set_size = screen_set.shape[0]
            active_set = np.empty(groups.shape[0], dtype=int)
            active_set[:active_set_size] = np.arange(active_set_size)

        else:
            lmda = warm_start.lmda
            lmda_max = warm_start.lmda_max
            screen_set = warm_start.screen_set
            screen_beta = warm_start.screen_beta
            screen_is_active = warm_start.screen_is_active
            screen_dual = warm_start.screen_dual
            active_set_size = warm_start.active_set_size
            active_set = warm_start.active_set

        solver_args["groups"] = groups
        solver_args["group_sizes"] = group_sizes
        solver_args["penalty"] = penalty
        solver_args["lmda"] = lmda
        solver_args["lmda_max"] = lmda_max
        solver_args["screen_set"] = screen_set
        solver_args["screen_beta"] = screen_beta
        solver_args["screen_is_active"] = screen_is_active
        solver_args["screen_dual"] = screen_dual
        solver_args["active_set_size"] = active_set_size
        solver_args["active_set"] = active_set

        # special gaussian case
        if is_gaussian_opt:
            y = glm.y
            weights = glm.weights
            if warm_start is None:
                ones = np.ones(n, dtype=dtype)
                X_means = np.empty(p, dtype=dtype)
                X.mul(ones, weights, X_means)
                y_off = y - offsets
                y_mean = np.sum(y_off * weights)
                yc = y_off
                if intercept:
                    yc = yc - y_mean
                y_var = np.sum(weights * yc ** 2)
                rsq = 0
                resid = yc
                resid_sum = np.sum(weights * resid)
                grad = np.empty(p, dtype=dtype)
                X.mul(resid, weights, grad)
            else:
                X_means = warm_start.X_means
                y_mean = warm_start.y_mean
                y_var = warm_start.y_var
                rsq = warm_start.rsq
                resid = warm_start.resid
                resid_sum = warm_start.resid_sum
                grad = warm_start.grad

            solver_args["X_means"] = X_means
            solver_args["y_mean"] = y_mean
            solver_args["y_var"] = y_var
            solver_args["rsq"] = rsq
            solver_args["resid"] = resid
            solver_args["resid_sum"] = resid_sum
            solver_args["grad"] = grad

            state = state_gaussian_naive(**solver_args)

        # GLM case
        else:
            if warm_start is None:
                ones = np.ones(n, dtype=dtype)
                beta0 = 0
                eta = offsets
                resid = np.empty(n, dtype=dtype)
                glm.gradient(eta, resid)
                grad = np.empty(p, dtype=dtype)
                X.mul(resid, ones, grad)
                loss_null = None
                loss_full = glm.loss_full()
            else:
                beta0 = warm_start.beta0
                eta = warm_start.eta
                resid = warm_start.resid
                grad = warm_start.grad
                loss_null = warm_start.loss_null
                loss_full = warm_start.loss_full

            solver_args["beta0"] = beta0
            solver_args["grad"] = grad
            solver_args["eta"] = eta
            solver_args["resid"] = resid
            solver_args["loss_null"] = loss_null
            solver_args["loss_full"] = loss_full
            state = state_glm_naive(**solver_args)

    if check_state:
        state.check(method="assert")

    return _solve(
        state=state,
        progress_bar=progress_bar,
        exit_cond=exit_cond,
    )



from .glm import (
    binomial,
    cox,
    gaussian,
    multigaussian,
    multinomial,
    poisson,
)
import numpy as np
import warnings 


def _sample_y(
    glm: str,
    eta: np.ndarray,
    beta: np.ndarray,
    rho: float =0,
    snr: float =1,
):
    n, K = eta.shape
    is_multi = "multi" in glm
    if not is_multi and K > 1:
        warnings.warn("Ignoring K and taking only first class response.")
        eta = eta[:, 0][:, None]
        K = 1

    if "gaussian" in glm:
        signal_scale = np.sqrt(rho * np.sum(beta) ** 2 + (1-rho) * np.sum(beta ** 2))
        noise_scale = signal_scale / np.sqrt(snr)
        y = eta + noise_scale * np.random.normal(0, 1, eta.shape)
        if not is_multi:
            y = y.ravel()
            return gaussian(y=y)
        return multigaussian(y=y)
    elif glm == "multinomial":
        signal_scale = np.sqrt(rho * np.sum(beta, axis=0) ** 2 + (1-rho) * np.sum(beta ** 2, axis=0))
        noise_scale = signal_scale / np.sqrt(snr)
        mu = np.empty((n, K), dtype=eta.dtype)
        glm = multinomial(y=np.zeros(eta.shape)) 
        glm.gradient(
            eta / noise_scale[None], 
            mu,
        )
        mu *= -K * n
        y = np.array([
            np.random.multinomial(1, m)
            for m in mu
        ], dtype=eta.dtype)
        return multinomial(y=y)
    elif glm == "cox":
        signal_scale = np.sqrt(rho * np.sum(beta) ** 2 + (1-rho) * np.sum(beta ** 2))
        noise_scale = signal_scale / np.sqrt(snr)
        eta = eta.ravel()
        s = np.round(np.random.exponential(1, n))
        t = 1 + s + np.round(np.exp(eta / noise_scale + np.random.normal(0, 1, n)))
        C = 1 + s + np.round(np.exp(np.random.normal(0, 1, n)))
        d = (t < C).astype(eta.dtype)
        t = np.minimum(t, C)
        return cox(
            start=s,
            stop=t,
            status=d,
        )
    else:
        signal_scale = np.sqrt(rho * np.sum(beta) ** 2 + (1-rho) * np.sum(beta ** 2))
        noise_scale = signal_scale / np.sqrt(snr)
        func_map = {
            "binomial": binomial,
            "poisson": poisson,
        }
        sample_map = {
            "binomial": lambda mu: np.random.binomial(1, mu),
            "poisson": lambda mu: np.random.poisson(mu),
        }
        glm_o = func_map[glm](y=np.zeros(n))
        eta = eta.ravel()
        mu = np.empty(eta.shape[0], dtype=eta.dtype)
        glm_o.gradient(eta / noise_scale, mu)
        mu *= -n
        y = sample_map[glm](mu).astype(eta.dtype)
        return func_map[glm](y=y)


def dense(
    n: int, 
    p: int, 
    G: int,
    *,
    K: int=1,
    glm: str ="gaussian",
    equal_groups=False,
    rho: float =0,
    sparsity: float =0.95,
    zero_penalty: float =0,
    snr: float = 1,
    seed: int =0,
):
    """Creates a dense dataset.

    - The groups and group sizes are generated randomly
      such that ``G`` groups are created and the sum of the group sizes is ``p``.
    - The data matrix ``X`` is generated from a normal distribution
      where each feature is equicorrelated with the other features by ``rho``.
    - The true coefficients :math:`\\beta` are such that ``sparsity`` proportion
      of the entries are set to :math:`0`.
    - The response ``y`` is generated from the GLM specified by ``glm``.
    - The penalty factors are by default set to ``np.sqrt(group_sizes)``,
      however if ``zero_penalty > 0``, a random set of penalties will be set to zero,
      in which case, ``penalty`` is rescaled such that the :math:`\\ell_2` norm squared equals ``p``.

    Parameters
    ----------
    n : int
        Number of data points.
    p : int
        Number of features.
    G : int
        Number of groups.
    K : int, optional
        Number of classes for multi-response GLMs.
        Default is ``1``.
    glm : str, optional
        GLM name.
        It must be one of the following:

            - ``"binomial"``
            - ``"cox"``
            - ``"gaussian"``
            - ``"multigaussian"``
            - ``"multinomial"``
            - ``"poisson"``

        Default is ``"gaussian"``.
    equal_groups : bool, optional
        If ``True``, group sizes are made as equal as possible.
        Default is ``False``.
    rho : float, optional
        Feature (equi)-correlation.
        Default is ``0`` so that the features are independent.
    sparsity : float, optional
        Proportion of :math:`\\beta` entries to be zeroed out.
        Default is ``0.95``.
    zero_penalty : float, optional
        Proportion of ``penalty`` entries to be zeroed out.
        Default is ``0``.
    snr : float, optional
        Signal-to-noise ratio.
        Default is ``1``.
    seed : int, optional
        Random seed.
        Default is ``0``.

    Returns
    -------
    data : dict
        A dictionary containing the generated data:
        
            - ``"X"``: feature matrix.
            - ``"y"``: response vector.
            - ``"groups"``: mapping of group index to the starting column index of ``X``.
            - ``"group_sizes"``: mapping of group index to the group size.
            - ``"penalty"``: penalty factor for each group index.
    """
    assert n >= 1
    assert p >= 1
    assert G >= 1
    assert rho >= -1 / (p-1)
    assert (0 <= sparsity) & (sparsity <= 1)
    assert (0 <= zero_penalty) & (zero_penalty <= 1)
    assert snr > 0
    assert seed >= 0

    np.random.seed(seed)

    # define groups
    if equal_groups:
        equal_group_size = p // G
        groups = equal_group_size * np.arange(G)
    else:
        groups = np.concatenate([
            [0],
            np.random.choice(np.arange(1, p), size=G-1, replace=False)
        ])
        groups = np.sort(groups).astype(int)
    group_sizes = np.concatenate([groups, [p]], dtype=int)
    group_sizes = group_sizes[1:] - group_sizes[:-1]

    penalty = np.sqrt(group_sizes)
    penalty[np.random.choice(G, int(zero_penalty * G), replace=False)] = 0
    penalty /= np.linalg.norm(penalty) / np.sqrt(p)

    # generate raw data
    X = np.random.normal(0, 1, (n, p))
    Z = np.random.normal(0, 1, n)
    X = np.sqrt(rho) * Z[:, None] + np.sqrt(1-rho) * X
    X = np.asfortranarray(X)

    beta = np.random.normal(0, 1, (p, K))
    beta_zero_indices = np.random.choice(p, int(sparsity * p), replace=False)
    beta_nnz_indices = np.array(list(set(np.arange(p)) - set(beta_zero_indices)))
    X_sub = X[:, beta_nnz_indices]
    beta_sub = beta[beta_nnz_indices]

    eta = X_sub @ beta_sub
    glm = _sample_y(
        glm=glm,
        eta=eta,
        beta=beta_sub,
        rho=rho,
        snr=snr,
    )

    return {
        "X": X, 
        "glm": glm,
        "groups": groups,
        "group_sizes": group_sizes,
        "penalty": penalty,
    }


def snp_unphased(
    n: int, 
    p: int, 
    *,
    K: int =1,
    glm: str ="gaussian",
    sparsity: float =0.95,
    missing_ratio: float =0.1,
    one_ratio: float =0.25,
    two_ratio: float =0.05,
    zero_penalty: float =0,
    snr: float =1,
    seed: int =0,
):
    """Creates a SNP unphased dataset.

    - This dataset is only used for lasso, so ``groups`` is simply each individual feature
      and ``group_sizes`` is a vector of ones.
    - The calldata matrix ``X`` has sparsity ratio ``1 - one_ratio - two_ratio``
      where ``one_ratio`` of the entries are randomly set to ``1``
      and ``two_ratio`` are randomly set to ``2``.
      The user only sees a masked version of ``X`` where
      ``missing_ratio`` of the entries are set to ``-9``.
    - The true coefficients :math:`\\beta` are such that ``sparsity`` proportion
      of the entries are set to :math:`0`.
    - The response ``y`` is generated from the GLM specified by ``glm``.
    - The penalty factors are by default set to ``np.sqrt(group_sizes)``,
      however if ``zero_penalty > 0``, a random set of penalties will be set to zero,
      in which case, ``penalty`` is rescaled such that the :math:`\\ell_2` norm squared is ``p``.

    Parameters
    ----------
    n : int
        Number of data points.
    p : int
        Number of SNPs.
    K : int, optional
        Number of classes for multi-response GLMs.
        Default is ``1``.
    glm : str, optional
        GLM name.
        It must be one of the following:

            - ``"binomial"``
            - ``"cox"``
            - ``"gaussian"``
            - ``"multigaussian"``
            - ``"multinomial"``
            - ``"poisson"``

        Default is ``"gaussian"``.
    sparsity : float, optional
        Proportion of :math:`\\beta` entries to be zeroed out.
        Default is ``0.95``.
    missing_ratio : float, optional
        Proportion of the entries of ``X`` that is set to ``-9`` (missing).
        Default is ``0.1``.
    one_ratio : float, optional
        Proportion of the entries of ``X`` that is set to ``1``.
        Default is ``0.25``.
    two_ratio : float, optional
        Proportion of the entries of ``X`` that is set to ``2``.
        Default is ``0.05``.
    zero_penalty : float, optional
        Proportion of ``penalty`` entries to be zeroed out.
        Default is ``0``.
    snr : float, optional
        Signal-to-noise ratio.
        Default is ``1``.
    seed : int, optional
        Random seed.
        Default is ``0``.

    Returns
    -------
    data : dict
        A dictionary containing the generated data:
        
            - ``"X"``: feature matrix.
            - ``"y"``: response vector.
            - ``"groups"``: mapping of group index to the starting column index of ``X``.
            - ``"group_sizes"``: mapping of group index to the group size.
            - ``"penalty"``: penalty factor for each group index.
    """
    assert n >= 1
    assert p >= 1
    assert sparsity >= 0 and sparsity <= 1
    assert missing_ratio >= 0 and missing_ratio <= 1
    assert one_ratio >= 0 and one_ratio <= 1
    assert two_ratio >= 0 and two_ratio <= 1
    assert zero_penalty >= 0 and zero_penalty <= 1
    assert snr > 0
    assert seed >= 0

    np.random.seed(seed)

    nnz_ratio = one_ratio + two_ratio
    one_ratio = one_ratio / nnz_ratio
    two_ratio = two_ratio / nnz_ratio
    nnz = int(nnz_ratio * n * p)
    nnz_indices = np.random.choice(n * p, nnz, replace=False)
    nnz_indices = np.random.permutation(nnz_indices)
    one_indices = nnz_indices[:int(one_ratio * nnz)]
    two_indices = nnz_indices[int(one_ratio * nnz):]
    X = np.zeros((n, p), dtype=np.int8)
    X.ravel()[one_indices] = 1
    X.ravel()[two_indices] = 2

    groups = np.arange(p)
    group_sizes = np.ones(p, dtype=int)

    penalty = np.sqrt(group_sizes)
    penalty[np.random.choice(p, int(zero_penalty * p), replace=False)] = 0
    penalty /= np.linalg.norm(penalty) / np.sqrt(p)

    beta = np.random.normal(0, 1, (p, K))
    beta_nnz_indices = np.random.choice(p, int((1-sparsity) * p), replace=False)
    X_sub = X[:, beta_nnz_indices]
    beta_sub = beta[beta_nnz_indices]

    eta = X_sub @ beta_sub 
    glm = _sample_y(
        glm=glm,
        eta=eta,
        beta=beta_sub,
        snr=snr,
    )

    n_missing = int(missing_ratio * n * p)
    X.ravel()[np.random.choice(n * p, n_missing, replace=False)] = -9

    return {
        "X": np.asfortranarray(X), 
        "glm": glm,
        "groups": groups,
        "group_sizes": group_sizes,
        "penalty": penalty,
    }


def snp_phased_ancestry(
    n: int, 
    s: int, 
    A: int,
    *,
    K: int =1,
    glm: str ="gaussian",
    sparsity: float =0.95,
    one_ratio: float =0.25,
    two_ratio: float =0.05,
    zero_penalty: float =0,
    snr: float =1,
    seed: int =0,
):
    """Creates a SNP phased, ancestry dataset.

    - The groups and group sizes are generated randomly
      such that ``G`` groups are created and the sum of the group sizes is ``p``.
    - The calldata matrix ``X`` is a phased version of a matrix with sparsity ratio 
      ``1 - one_ratio - two_ratio``
      where ``one_ratio`` of the entries are randomly set to ``1``
      and ``two_ratio`` are randomly set to ``2``.
    - The ancestry matrix randomly generates integers in the range ``[0, A)``.
    - The true coefficients :math:`\\beta` is such that ``sparsity`` proportion of the entries are set to :math:`0`.
    - The response ``y`` is generated from the GLM specified by ``glm``.
    - The penalty factors are by default set to ``np.sqrt(group_sizes)``,
      however if ``zero_penalty > 0``, a random set of penalties will be set to zero,
      in which case, ``penalty`` is rescaled such that the :math:`\\ell_2` norm squared is ``p``.

    Parameters
    ----------
    n : int
        Number of data points.
    s : int
        Number of SNPs.
    A : int
        Number of ancestries.
    K : int, optional
        Number of classes for multi-response GLMs.
        Default is ``1``.
    glm : str, optional
        GLM name.
        It must be one of the following:

            - ``"binomial"``
            - ``"cox"``
            - ``"gaussian"``
            - ``"multigaussian"``
            - ``"multinomial"``
            - ``"poisson"``

        Default is ``"gaussian"``.
    sparsity : float, optional
        Proportion of :math:`\\beta` entries to be zeroed out.
        Default is ``0.95``.
    one_ratio : float, optional
        Proportion of the entries of ``X`` that is set to ``1``.
        Default is ``0.25``.
    two_ratio : float, optional
        Proportion of the entries of ``X`` that is set to ``2``.
        Default is ``0.05``.
    zero_penalty : float, optional
        Proportion of ``penalty`` entries to be zeroed out.
        Default is ``0``.
    snr : float, optional
        Signal-to-noise ratio.
        Default is ``1``.
    seed : int, optional
        Random seed.
        Default is ``0``.

    Returns
    -------
    data : dict
        A dictionary containing the generated data:
        
            - ``"X"``: feature matrix.
            - ``"ancestries"``: ancestry label of the same shape as ``X``.
            - ``"y"``: response vector.
            - ``"groups"``: mapping of group index to the starting column index of ``X``.
            - ``"group_sizes"``: mapping of group index to the group size.
            - ``"penalty"``: penalty factor for each group index.
    """
    assert n >= 1
    assert s >= 1
    assert A >= 1
    assert sparsity >= 0 and sparsity <= 1
    assert one_ratio >= 0 and one_ratio <= 1
    assert two_ratio >= 0 and two_ratio <= 1
    assert zero_penalty >= 0 and zero_penalty <= 1
    assert snr > 0
    assert seed >= 0

    np.random.seed(seed)

    nnz_ratio = one_ratio + two_ratio
    one_ratio = one_ratio / nnz_ratio
    two_ratio = two_ratio / nnz_ratio
    nnz = int(nnz_ratio * n * s)
    nnz_indices = np.random.choice(n * s, nnz, replace=False)
    nnz_indices = np.random.permutation(nnz_indices)
    one_indices = nnz_indices[:int(one_ratio * nnz)]
    two_indices = nnz_indices[int(one_ratio * nnz):]
    one_indices = 2 * one_indices + np.random.binomial(1, 0.5, len(one_indices))
    two_indices *= 2
    X = np.zeros((n, 2 * s), dtype=np.int8)
    X.ravel()[one_indices] = 1
    X.ravel()[two_indices] = 1
    X.ravel()[two_indices + 1] = 1
    ancestries = np.zeros((n, 2 * s), dtype=np.int8)
    ancestries.ravel()[one_indices] = np.random.choice(A, len(one_indices), replace=True)
    ancestries.ravel()[two_indices] = np.random.choice(A, len(two_indices), replace=True)
    ancestries.ravel()[two_indices + 1] = np.random.choice(A, len(two_indices), replace=True)

    groups = A * np.arange(s)
    group_sizes = np.full(s, A, dtype=int)

    penalty = np.sqrt(group_sizes)
    penalty[np.random.choice(s, int(zero_penalty * s), replace=False)] = 0
    penalty /= np.linalg.norm(penalty) / np.sqrt(s * A)

    beta = np.random.normal(0, 1, (2 * s, K))
    beta_nnz_indices = np.random.choice(2 * s, int((1-sparsity) * s * 2), replace=False)
    X_sub = X[:, beta_nnz_indices]
    beta_sub = beta[beta_nnz_indices]

    eta = X_sub @ beta_sub 
    glm = _sample_y(
        glm=glm,
        eta=eta,
        beta=beta_sub,
        snr=snr,
    )

    return {
        "X": np.asfortranarray(X), 
        "ancestries": np.asfortranarray(ancestries),
        "glm": glm,
        "groups": groups,
        "group_sizes": group_sizes,
        "penalty": penalty,
    }



from . import adelie_core as core
from .adelie_core.glm import (
    GlmBase64,
    GlmBase32,
    GlmMultiBase64,
    GlmMultiBase32,
)
from typing import Union
import numpy as np
import warnings


def _coerce_dtype(y, dtype):
    dtype_map = {
        np.dtype("float32"): np.float32,
        np.dtype("float64"): np.float64,
    }
    valid_dtypes = list(dtype_map.keys())
    y = np.array(y, order="C")
    if dtype is None:
        if not (y.dtype in valid_dtypes):
            raise RuntimeError(
                "y must have an underlying type of np.float32 or np.float64, "
                "or dtype must be explicitly specified."
            )
        dtype = dtype_map[y.dtype]
    else:
        if not (dtype in valid_dtypes):
            raise RuntimeError("dtype must be either np.float32 or np.float64.")
    y = y.astype(dtype)
    return y, dtype


class glm_base:
    """Base wrapper GLM class.

    All Python wrapper classes for core GLM classes must inherit from this class.
    The purpose of this class is to expose extra member interface
    that are more easily written in Python (and not speed-critical).
    """
    def __init__(self, y, weights, core_base, dtype):
        self.core_base = core_base
        self.y = np.array(y, copy=True, dtype=dtype)
        self.dtype = dtype
        if len(y.shape) != 1:
            raise RuntimeError("y must be 1-dimensional.")
        n = y.shape[0]
        if not (weights is None):
            if weights.shape != (n,):
                raise RuntimeError("y and weights must have same length.")
            weights_sum = np.sum(weights)
            if not np.allclose(weights_sum, 1):
                weights = weights / weights_sum
        else:
            weights = np.full(n, 1/n, dtype=dtype)
        self.weights = np.array(weights, copy=True, dtype=dtype)


class multiglm_base:
    """Base wrapper multi-response GLM class.

    All Python wrapper classes for core multi-response GLM classes must inherit from this class.
    The purpose of this class is to expose extra member interface
    that are more easily written in Python (and not speed-critical).
    """
    def __init__(self, y, weights, core_base, dtype):
        self.core_base = core_base
        self.y = np.array(y, copy=True, dtype=dtype)
        self.dtype = dtype
        if len(y.shape) != 2:
            raise RuntimeError("y must be 2-dimensional.")
        n = y.shape[0]
        if not (weights is None):
            if weights.shape != (n,):
                raise RuntimeError("y rows and weights must have same length.")
            weights_sum = np.sum(weights)
            if not np.allclose(weights_sum, 1):
                weights = weights / np.sum(weights)
        else:
            weights = np.full(n, 1/n, dtype=dtype)
        self.weights = np.array(weights, copy=True, dtype=dtype)


def binomial(
    y: np.ndarray,
    *,
    weights: np.ndarray =None,
    link: str ="logit",
    dtype: Union[np.float32, np.float64] =None,
):
    """Creates a Binomial GLM family object.

    The Binomial GLM family with the logit link function 
    specifies the loss function as:

    .. math::
        \\begin{align*}
            \\ell(\\eta)
            =
            \\sum\\limits_{i=1}^n w_i \\left(
                -y_i \\eta_i + \\log(1 + e^{\\eta_i})
            \\right)
        \\end{align*}

    The Binomial GLM family with the probit link function 
    specifies the loss function as:

    .. math::
        \\begin{align*}
            \\ell(\\eta)
            =
            -\\sum\\limits_{i=1}^n w_i \\left(
                y_i \\log(\\Phi(\\eta_i)) + (1-y_i) \\log(1-\\Phi(\\eta_i))
            \\right)
        \\end{align*}

    where :math:`\\Phi` is the standard normal CDF.

    We assume that :math:`y_i \\in [0,1]`.

    Parameters
    ----------
    y : (n,) np.ndarray 
        Response vector :math:`y`.
    weights : (n,) np.ndarray, optional
        Observation weights :math:`W`.
        Weights are normalized such that they sum to ``1``.
        Default is ``None``, in which case, it is set to ``np.full(n, 1/n)``.
    link : str, optional
        The link function type.
        It must be one of the following:

            - ``"logit"``: the logit link function.
            - ``"probit"``: the probit link function.

        Default is ``"logit"``.
    dtype : Union[np.float32, np.float64], optional
        The underlying data type.
        If ``None``, it is inferred from ``y``,
        in which case ``y`` must have an underlying data type of
        ``np.float32`` or ``np.float64``.
        Default is ``None``.

    Returns
    -------
    glm
        Binomial GLM object.

    See Also
    --------
    adelie.adelie_core.glm.GlmBinomialLogit64
    adelie.adelie_core.glm.GlmBinomialProbit64
    """
    dispatcher = {
        "logit": {
            np.float64: core.glm.GlmBinomialLogit64,
            np.float32: core.glm.GlmBinomialLogit32,
        },
        "probit": {
            np.float64: core.glm.GlmBinomialProbit64,
            np.float32: core.glm.GlmBinomialProbit32,
        },
    }

    y, dtype = _coerce_dtype(y, dtype)

    core_base = dispatcher[link][dtype]

    class _binomial(glm_base, core_base):
        def __init__(self):
            glm_base.__init__(self, y, weights, core_base, dtype)
            core_base.__init__(self, self.y, self.weights)

        def reweight(self, weights=None):
            weights = self.weights if weights is None else weights
            return binomial(y=y, weights=weights, dtype=dtype)

    return _binomial()


def cox(
    start: np.ndarray,
    stop: np.ndarray,
    status: np.ndarray,
    *,
    weights: np.ndarray =None,
    tie_method: str ="efron",
    dtype: Union[np.float32, np.float64] =None,
):
    """Creates a Cox GLM family object.

    The Cox GLM family specifies the loss function as:

    .. math::
        \\begin{align*}
            \\ell(\\eta)
            &=
            -\\sum\\limits_{i=1}^n w_i \\delta_i \\eta_i
            +\\sum\\limits_{i=1}^n \\overline{w}_i \\delta_i A_i(\\eta)
            \\\\
            A_i(\\eta)
            &=
            \\log\\left(
                \\sum\\limits_{k \\in R(t_i)} w_k e^{\\eta_k}
                -
                \\sigma_i \\sum\\limits_{k \\in H(t_i)} w_k e^{\\eta_k}
            \\right)
        \\end{align*}

    where
    
    .. math::
        \\begin{align*}
            R(u)
            &=
            \\{i : u \\in (s_i, t_i]\\}
            \\\\
            H(u)
            &=
            \\{i : t_i = u, \\delta_i = 1\\}
            \\\\
            \\overline{w}_i 
            &= 
            \\frac{\\sum_{k \\in H(t_i)} w_k}{\\sum_{k \\in H(t_i)} 1_{w_k > 0}} 
            1_{\\delta_i = 1, w_i > 0}
        \\end{align*}

    Here,
    :math:`\\delta` is the status (``1`` for event, ``0`` for censored) vector,
    :math:`s` is the vector of start times,
    :math:`t` is the vector of stop times,
    :math:`R(u)` is the at-risk set at time :math:`u`,
    :math:`H(u)` is the set of ties at event time :math:`u`,
    :math:`\\overline{w}` is the vector of average weights within ties with positive weights,
    :math:`\\sigma` is the correction scale for tie-breaks, 
    which is determined by the type of correction method (Breslow or Efron).
    Note that :math:`\\overline{w}_i` and :math:`A_i(\\eta)` are only well-defined 
    whenever :math:`\\delta_i=1`, which is not an issue in the computation of :math:`\\ell(\\eta)`.

    Parameters
    ----------
    start : (n,) np.ndarray
        Start time vector :math:`s`.
    stop : (n,) np.ndarray
        Stop time vector :math:`t`.
    status : (n,) np.ndarray 
        Status vector :math:`\\delta`.
    weights : (n,) np.ndarray, optional
        Observation weights :math:`W`.
        Weights are normalized such that they sum to ``1``.
        Default is ``None``, in which case, it is set to ``np.full(n, 1/n)``.
    tie_method : str, optional
        The tie-breaking method that determines the scales :math:`\\sigma`.
        It must be one of the following:

            - ``"efron"``
            - ``"breslow"``

        Default is ``"efron"``.
    dtype : Union[np.float32, np.float64], optional
        The underlying data type.
        If ``None``, it is inferred from ``status``,
        in which case ``status`` must have an underlying data type of
        ``np.float32`` or ``np.float64``.
        Default is ``None``.

    Returns
    -------
    glm
        Cox GLM object.

    See Also
    --------
    adelie.adelie_core.glm.GlmCox64
    """
    dispatcher = {
        np.float64: core.glm.GlmCox64,
        np.float32: core.glm.GlmCox32,
    }

    status, dtype = _coerce_dtype(status, dtype)

    core_base = dispatcher[dtype]

    class _cox(glm_base, core_base):
        def __init__(self):
            self.start = np.array(start, copy=True, dtype=dtype)
            self.stop = np.array(stop, copy=True, dtype=dtype)
            glm_base.__init__(self, status, weights, core_base, dtype)
            self.status = self.y
            self.tie_method = tie_method
            core_base.__init__(
                self, 
                self.start, 
                self.stop, 
                self.status, 
                self.weights, 
                self.tie_method,
            )

        def reweight(self, weights=None):
            weights = self.weights if weights is None else weights
            return cox(
                start=start, 
                stop=stop,
                status=status,
                weights=weights, 
                tie_method=tie_method,
                dtype=dtype,
            )

    return _cox()


def gaussian(
    y: np.ndarray,
    *,
    weights: np.ndarray =None,
    dtype: Union[np.float32, np.float64] =None,
    opt: bool =True,
):
    """Creates a Gaussian GLM family object.

    The Gaussian GLM family specifies the loss function as:

    .. math::
        \\begin{align*}
            \\ell(\\eta)
            =
            \\sum\\limits_{i=1}^n w_i \\left(
                -y_i \\eta_i + \\frac{\\eta_i^2}{2}
            \\right) 
        \\end{align*}

    Parameters
    ----------
    y : (n,) np.ndarray 
        Response vector :math:`y`.
    weights : (n,) np.ndarray, optional
        Observation weights :math:`W`.
        Weights are normalized such that they sum to ``1``.
        Default is ``None``, in which case, it is set to ``np.full(n, 1/n)``.
    dtype : Union[np.float32, np.float64], optional
        The underlying data type.
        If ``None``, it is inferred from ``y``,
        in which case ``y`` must have an underlying data type of
        ``np.float32`` or ``np.float64``.
        Default is ``None``.
    opt : bool, optional
        If ``True``, an optimized routine is used when passed into ``adelie.grpnet``.
        Otherwise, a general routine with IRLS is used.
        This flag is mainly for developers for testing purposes.
        We advise users to use the default value.
        Default is ``True``.

    Returns
    -------
    glm
        Gaussian GLM object.

    See Also
    --------
    adelie.adelie_core.glm.GlmGaussian64
    """
    dispatcher = {
        np.float64: core.glm.GlmGaussian64,
        np.float32: core.glm.GlmGaussian32,
    }

    y, dtype = _coerce_dtype(y, dtype)

    core_base = dispatcher[dtype]

    class _gaussian(glm_base, core_base):
        def __init__(self):
            self.opt = opt
            glm_base.__init__(self, y, weights, core_base, dtype)
            core_base.__init__(self, self.y, self.weights)

        def reweight(self, weights=None):
            weights = self.weights if weights is None else weights
            return gaussian(y=y, weights=weights, dtype=dtype, opt=opt)

    return _gaussian()


def multigaussian(
    y: np.ndarray,
    *,
    weights: np.ndarray =None,
    dtype: Union[np.float32, np.float64] =None,
    opt: bool =True,
):
    """Creates a MultiGaussian GLM family object.

    The MultiGaussian GLM family specifies the loss function as:

    .. math::
        \\begin{align*}
            \\ell(\\eta)
            =
            \\frac{1}{K}
            \\sum\\limits_{i=1}^n 
            w_{i} \\left(
                -\\sum\\limits_{k=1}^K y_{ik} \\eta_{ik} 
                +\\frac{\\|\\eta_{i\\cdot}\\|^2}{2}
            \\right)
        \\end{align*}

    Parameters
    ----------
    y : (n, K) np.ndarray 
        Response matrix :math:`y`.
    weights : (n,) np.ndarray, optional
        Observation weights :math:`W`.
        Weights are normalized such that they sum to ``1``.
        Default is ``None``, in which case, it is set to ``np.full(n, 1/n)``.
    dtype : Union[np.float32, np.float64], optional
        The underlying data type.
        If ``None``, it is inferred from ``y``,
        in which case ``y`` must have an underlying data type of
        ``np.float32`` or ``np.float64``.
        Default is ``None``.
    opt : bool, optional
        If ``True``, an optimized routine is used when passed into ``adelie.grpnet``.
        Otherwise, a general routine with IRLS is used.
        This flag is mainly for developers for testing purposes.
        We advise users to use the default value.
        Default is ``True``.

    Returns
    -------
    glm
        MultiGaussian GLM object.

    See Also
    --------
    adelie.adelie_core.glm.GlmMultiGaussian64
    """
    dispatcher = {
        np.float64: core.glm.GlmMultiGaussian64,
        np.float32: core.glm.GlmMultiGaussian32,
    }

    y, dtype = _coerce_dtype(y, dtype)

    core_base = dispatcher[dtype]

    class _multigaussian(multiglm_base, core_base):
        def __init__(self):
            self.opt = opt
            multiglm_base.__init__(self, y, weights, core_base, dtype)
            core_base.__init__(self, self.y, self.weights)

        def reweight(self, weights=None):
            weights = self.weights if weights is None else weights
            return multigaussian(y=y, weights=weights, dtype=dtype, opt=opt)

    return _multigaussian()


def multinomial(
    y: np.ndarray,
    *,
    weights: np.ndarray =None,
    dtype: Union[np.float32, np.float64] =None,
):
    """Creates a Multinomial GLM family object.

    The Multinomial GLM family specifies the loss function as:

    .. math::
        \\begin{align*}
            \\ell(\\eta)
            =
            \\frac{1}{K}
            \\sum\\limits_{i=1}^n 
            w_i
            \\left(
            -\\sum\\limits_{k=1}^{K} y_{ik} \\eta_{ik} 
            + \\log\\left(
                \\sum\\limits_{\\ell=1}^{K} e^{\\eta_{i\\ell}}
            \\right)
            \\right)
        \\end{align*}

    We assume that every :math:`y_{ik} \\in [0,1]` and
    for each fixed :math:`i`, :math:`\\sum_{k=1}^K y_{ik} = 1`.

    .. note::
        The ``hessian`` method computes :math:`2 \\mathrm{diag}(\\nabla^2 \\ell(\\eta))`
        as the diagonal majorization.

    Parameters
    ----------
    y : (n, K) np.ndarray 
        Response matrix :math:`y`.
    weights : (n,) np.ndarray, optional
        Observation weights :math:`W`.
        Weights are normalized such that they sum to ``1``.
        Default is ``None``, in which case, it is set to ``np.full(n, 1/n)``.
    dtype : Union[np.float32, np.float64], optional
        The underlying data type.
        If ``None``, it is inferred from ``y``,
        in which case ``y`` must have an underlying data type of
        ``np.float32`` or ``np.float64``.
        Default is ``None``.

    Returns
    -------
    glm
        Multinomial GLM object.

    See Also
    --------
    adelie.adelie_core.glm.GlmMultinomial64
    """
    dispatcher = {
        np.float64: core.glm.GlmMultinomial64,
        np.float32: core.glm.GlmMultinomial32,
    }

    y, dtype = _coerce_dtype(y, dtype)

    core_base = dispatcher[dtype]

    class _multinomial(multiglm_base, core_base):
        def __init__(self):
            multiglm_base.__init__(self, y, weights, core_base, dtype)
            core_base.__init__(self, self.y, self.weights)

        def reweight(self, weights=None):
            weights = self.weights if weights is None else weights
            return multinomial(y=y, weights=weights, dtype=dtype)

    return _multinomial()


def poisson(
    y: np.ndarray,
    *,
    weights: np.ndarray =None,
    dtype: Union[np.float32, np.float64] =None,
):
    """Creates a Poisson GLM family object.

    The Poisson GLM family specifies the loss function as:

    .. math::
        \\begin{align*}
            \\ell(\\eta)
            =
            \\sum\\limits_{i=1}^n w_i \\left(
                -y_i \\eta_i + e^{\\eta_i}
            \\right) 
        \\end{align*}

    We assume that :math:`y_i \\in \\mathbb{N}_0`.

    Parameters
    ----------
    y : (n,) np.ndarray 
        Response vector :math:`y`.
    weights : (n,) np.ndarray, optional
        Observation weights :math:`W`.
        Weights are normalized such that they sum to ``1``.
        Default is ``None``, in which case, it is set to ``np.full(n, 1/n)``.
    dtype : Union[np.float32, np.float64], optional
        The underlying data type.
        If ``None``, it is inferred from ``y``,
        in which case ``y`` must have an underlying data type of
        ``np.float32`` or ``np.float64``.
        Default is ``None``.

    Returns
    -------
    glm
        Poisson GLM object.

    See Also
    --------
    adelie.adelie_core.glm.GlmPoisson64
    """
    dispatcher = {
        np.float64: core.glm.GlmPoisson64,
        np.float32: core.glm.GlmPoisson32,
    }

    y, dtype = _coerce_dtype(y, dtype)

    core_base = dispatcher[dtype]

    class _poisson(glm_base, core_base):
        def __init__(self):
            glm_base.__init__(self, y, weights, core_base, dtype)
            core_base.__init__(self, self.y, self.weights)

        def reweight(self, weights=None):
            weights = self.weights if weights is None else weights
            return poisson(y=y, weights=weights, dtype=dtype)

    return _poisson()



from .adelie_core import io as core_io
from typing import Union
import numpy as np


class snp_phased_ancestry(core_io.IOSNPPhasedAncestry):
    """IO handler for SNP phased, ancestry matrix.

    A SNP phased, ancestry matrix is a matrix 
    that combines a (phased) calldata and local ancestry information.

    Let :math:`X \\in \\mathbb{R}^{n \\times sA}` denote such a matrix
    where
    :math:`n` is the number of samples,
    :math:`s` is the number of SNPs,
    and :math:`A` is the number of ancestries.
    Every :math:`A` (contiguous) columns is called an
    *ancestry block* corresponding to a single SNP
    with the same structure described next.
    Let :math:`H \\in \\mathbb{R}^{n \\times A}` denote any such ancestry block for some SNP.
    Then, :math:`H = H_0 + H_1` where :math:`H_k \\in \\mathbb{R}^{n \\times A}`
    represents the phased calldata marked by the ancestry indicator
    for each of the two haplotypes of a SNP, that is,

    .. math::
        \\begin{align*}
            H_k
            =
            \\begin{bmatrix}
                \\unicode{x2014} & \\delta^k_1 \\cdot e_{a^k_1}^\\top & \\unicode{x2014} \\\\
                \\unicode{x2014} & \\delta^k_2 \\cdot e_{a^k_2}^\\top & \\unicode{x2014} \\\\
                \\vdots & \\vdots & \\vdots \\\\
                \\unicode{x2014} & \\delta^k_n \\cdot e_{a^k_n}^\\top & \\unicode{x2014} \\\\
            \\end{bmatrix}
        \\end{align*}

    where for each individual :math:`i` and haplotype :math:`k`,
    :math:`\\delta^k_i \\in \\{0,1\\}` is :math:`1` 
    if and only if there is a mutation 
    and :math:`a^k_i \\in \\{1,\\ldots,A\\}` is 
    the ancestry labeling.
    Here, :math:`e_j \\in \\mathbb{R}^A` is the :math:`j` th standard basis vector.
         
    Parameters
    ----------
    filename : str
        File name containing the SNP data in ``.snpdat`` format.
    read_mode : str, optional
        Reading mode of the SNP data. 
        It must be one of the following:

            - ``"file"``: reads the file using standard file IO.
              This method is the most general and portable,
              however, with large files, it is the slowest option.
            - ``"mmap"``: reads the file using mmap.
              This method is only supported on Linux and MacOS.
              It is the most efficient way to read large files.

        Default is ``"file"``.
    """
    def __init__(
        self,
        filename: str,
        read_mode: str ="file",
    ):
        core_io.IOSNPPhasedAncestry.__init__(self, filename, read_mode)

    def write(
        self, 
        calldata: np.ndarray,
        ancestries: np.ndarray,
        A: int,
        n_threads: int =1,
    ):
        """Writes a dense SNP phased, ancestry matrix to the file in ``.snpdat`` format.

        .. note::
            The calldata and ancestries matrices must not contain
            any missing values.

        Parameters
        ----------
        calldata : (n, 2*s) np.ndarray
            SNP phased calldata in dense format.
            ``calldata[i, 2*j+k]`` is the data for individual ``i``, SNP ``j``, and haplotype ``k``.
            It must only contain values in :math:`\\{0,1\\}`.
        ancestries : (n, 2*s) np.ndarray
            Local ancestry information in dense format.
            ``ancestries[i, 2*j+k]`` is the ancestry for individual ``i``, SNP ``j``, and haplotype ``k``.
            It must only contain values in :math:`\\{0,\\ldots, A-1\\}`.
        A : int
            Number of ancestries.
        n_threads : int, optional
            Number of threads.
            Default is ``1``.

        Returns
        -------
        total_bytes : int
            Number of bytes written.
        benchmark : dict
            Dictionary of benchmark timings for each step of the serializer.
        """
        (
            total_bytes, 
            benchmark, 
            error,
        ) = core_io.IOSNPPhasedAncestry.write(self, calldata, ancestries, A, n_threads)
        if error != "":
            raise RuntimeError(error)
        return total_bytes, benchmark


class snp_unphased(core_io.IOSNPUnphased):
    """IO handler for SNP unphased matrix.

    A SNP unphased matrix is a matrix that contains values in the set ``{0, 1, 2, NA}``
    where ``NA`` indicates a missing value.
    Typically, ``NA`` is encoded as ``-9``, but for more generality
    we assume *any* negative value is equivalent to ``NA``.

    Parameters
    ----------
    filename : str
        File name to either read or write the SNP unphased matrix in ``.snpdat`` format.
    read_mode : str, optional
        Reading mode of the file ``filename``. 
        It must be one of the following:

            - ``"file"``: reads the file using standard file IO.
              This method is the most general and portable method,
              however, with large files, it is the slowest one.
            - ``"mmap"``: reads the file using mmap.
              This method is only supported on Linux and MacOS.
              It is the most efficient way to read large files.

        Default is ``"file"``.
    """
    def __init__(
        self,
        filename: str,
        read_mode: str ="file",
    ):
        core_io.IOSNPUnphased.__init__(self, filename, read_mode)

    def write(
        self, 
        calldata: np.ndarray,
        impute_method: Union[str, np.ndarray] ="mean",
        n_threads: int =1,
    ):
        """Writes a dense SNP unphased matrix to the file in ``.snpdat`` format.

        Parameters
        ----------
        calldata : (n, p) np.ndarray
            SNP unphased matrix in dense format.
        impute_method : Union[str, np.ndarray], optional
            Impute method for missing values. 
            It must be one of the following:

                - ``"mean"``: mean-imputation. Missing values in column ``j`` of ``calldata`` are replaced with
                  the mean of column ``j`` where the mean is computed using the non-missing values.
                  If every value is missing, we impute with ``0``.
                - ``np.ndarray``: user-specified vector of imputed values for each column of ``calldata``.
                
            Default is ``"mean"``.
        n_threads : int, optional
            Number of threads.
            Default is ``1``.

        Returns
        -------
        total_bytes : int
            Number of bytes written.
        benchmark : dict
            Dictionary of benchmark timings for each step of the serializer.
        """
        if isinstance(impute_method, str):
            p = calldata.shape[1]
            impute = np.empty(p)
        elif isinstance(impute_method, np.ndarray):
            impute = impute_method
            impute_method = "user"
        else:
            raise ValueError("impute_method must be a valid option.")
        (
            total_bytes, 
            benchmark, 
            error,
        ) = core_io.IOSNPUnphased.write(self, calldata, impute_method, impute, n_threads)
        if error != "":
            raise RuntimeError(error)
        return total_bytes, benchmark



from . import adelie_core as core
from .glm import (
    GlmBase32,
    GlmBase64,
    GlmMultiBase32,
    GlmMultiBase64,
)
from . import logger
from . import matrix
from .matrix import (
    MatrixNaiveBase32,
    MatrixNaiveBase64,
)
from IPython.display import HTML
from itertools import cycle
from scipy.sparse import csr_matrix
from typing import Union
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as animation


def predict(
    X: Union[MatrixNaiveBase32, MatrixNaiveBase64, np.ndarray],
    betas: Union[np.ndarray, csr_matrix],
    intercepts: np.ndarray,
    *,
    offsets: np.ndarray =None,
    n_threads: int =1,
):
    """Computes the linear predictions.

    The single-response linear prediction is given by
    
    .. math::
        \\begin{align*}
            \\hat{\\eta} = X\\beta + \\beta_0 \\mathbf{1} + \\eta^0
        \\end{align*}

    The multi-response linear prediction is given by

    .. math::
        \\begin{align*}
            \\hat{\\eta} = 
            (X\\otimes I_K) \\beta + 
            (\\mathbf{1}\\otimes I_k) \\beta_0 + 
            \\eta^0
        \\end{align*}

    The single or multi-response is detected based on the shape of ``intercepts``.
    If ``intercepts`` one-dimensional, we assume single-response.
    Otherwise, we assume multi-response.

    Parameters
    ----------
    X : (n, p) Union[MatrixNaiveBase32, MatrixNaiveBase64, np.ndarray]
        Feature matrix.
        It is typically one of the matrices defined in ``adelie.matrix`` submodule.
    betas : (L, p) or (L, p*K) Union[np.ndarray, scipy.sparse.csr_matrix]
        Coefficient vectors :math:`\\beta`.
    intercepts : (L,) or (L, K) np.ndarray
        Intercepts :math:`\\beta_0`.
    offsets : (n,) or (n, K) np.ndarray, optional
        Observation offsets :math:`\\eta^0`.
        Default is ``None``, in which case, it is set to 
        ``np.zeros(n)`` if ``y`` is single-response
        and ``np.zeros((n, K))`` if multi-response.
    n_threads : int, optional
        Number of threads.
        Default is ``1``.

    Returns
    -------
    linear_preds : (L, n) or (L, n, K) np.ndarray
        Linear predictions.

    See Also
    --------
    adelie.adelie_core.glm.GlmBase64
    adelie.adelie_core.glm.GlmMultiBase64
    """
    is_multi = len(intercepts.shape) == 2
    if is_multi:
        K = intercepts.shape[1]
        X = matrix.kronecker_eye(X, K, n_threads=n_threads)
        n = X.rows() // K
        y_shape = (n, K)
    else:
        if isinstance(X, np.ndarray):
            X = matrix.dense(X, method="naive", n_threads=n_threads)
        n = X.rows()
        y_shape = (n,)

    if offsets is None:
        offsets = np.zeros(y_shape)

    L = betas.shape[0]

    etas = np.zeros((L,) + y_shape, order="C")
    if isinstance(betas, np.ndarray):
        for i in range(etas.shape[0]):
            X.btmul(0, X.cols(), betas[i], etas[i].ravel())
    elif isinstance(betas, csr_matrix):
        X.sp_btmul(betas, etas.reshape((L, -1))) 
    else:
        raise RuntimeError("beta is not one of np.ndarray or scipy.sparse.csr_matrix.")
    etas += intercepts[:, None] + offsets

    return etas


def objective(
    X: Union[matrix.MatrixNaiveBase32, matrix.MatrixNaiveBase64, np.ndarray], 
    glm: Union[GlmBase32, GlmBase64, GlmMultiBase32, GlmMultiBase64],
    betas: Union[np.ndarray, csr_matrix], 
    intercepts: np.ndarray,
    lmdas: np.ndarray, 
    *,
    groups: np.ndarray =None, 
    alpha: float =1, 
    penalty: np.ndarray =None,
    offsets: np.ndarray =None,
    relative: bool =True,
    add_penalty: bool =True,
    n_threads: int =1,
):
    """Computes the group elastic net objective.

    See ``adelie.solver.grpnet`` for details.

    Parameters
    ----------
    X : (n, p) Union[adelie.matrix.MatrixNaiveBase32, adelie.matrix.MatrixNaiveBase64]
        Feature matrix :math:`X`.
        It is typically one of the matrices defined in ``adelie.matrix`` submodule or ``np.ndarray``.
    glm : Union[glm.GlmBase32, glm.GlmBase64, glm.GlmMultiBase32, glm.GlmMultiBase64] 
        GLM object.
        It is typically one of the GLM classes defined in ``adelie.glm`` submodule.
    betas : (L, p) or (L, p*K) Union[np.ndarray, scipy.sparse.csr_matrix]
        Coefficient vectors :math:`\\beta`.
    intercepts : (L,) or (L, K) np.ndarray
        Intercepts :math:`\\beta_0`.
    lmdas : (L,) np.ndarray 
        Regularization parameters :math:`\\lambda`.
        It is only used when ``add_penalty=True``.
        Otherwise, the user may pass ``None``.
    groups : (G,) np.ndarray, optional
        List of starting indices to each group where `G` is the number of groups.
        ``groups[i]`` is the starting index of the ``i`` th group. 
        If ``glm`` is multi-response type, then we only allow two types of groupings:

            - ``"grouped"``: coefficients for each predictor is grouped across the classes.
            - ``"ungrouped"``: every coefficient is its own group.

        It is only used when ``add_penalty=True``.
        Default is ``None``, in which case it is set to
        ``np.arange(p)`` if ``y`` is single-response
        and ``"grouped"`` if multi-response.
    alpha : float, optional
        Elastic net parameter :math:`\\alpha`.
        It must be in the range :math:`[0,1]`.
        It is only used when ``add_penalty=True``.
        Otherwise, the user may pass ``None``.
        Default is ``1``.
    penalty : (G,) np.ndarray, optional
        Penalty factor for each group in the same order as ``groups``.
        It must be a non-negative vector.
        It is only used when ``add_penalty=True``.
        Default is ``None``, in which case, it is set to ``np.sqrt(group_sizes)``.
    offsets : (n,) or (n, K) np.ndarray, optional
        Observation offsets :math:`\\eta^0`.
        Default is ``None``, in which case, it is set to 
        ``np.zeros(n)`` if ``y`` is single-response
        and ``np.zeros((n, K))`` if multi-response.
    relative : bool, optional
        If ``True``, then the full loss, :math:`\\ell(\\eta^\\star)`, is computed at the saturated model
        and the difference :math:`\\ell(\\eta)-\\ell(\\eta^\\star)` is provided,
        which will always be non-negative.
        This effectively computes loss *relative* to the saturated model.
        Default is ``True``.
    add_penalty : bool, optional
        If ``False``, the regularization term is removed 
        so that only the loss part is calculated. 
        Default is ``True``.
    n_threads : int, optional
        Number of threads.
        Default is ``1``.
    
    Returns
    -------
    obj : (L,) np.ndarray
        Group elastic net objectives.

    See Also
    --------
    adelie.solver.grpnet
    """
    X_raw = X
    y = glm.y
    if glm.is_multi:
        K = y.shape[1]
        X = matrix.kronecker_eye(X, K, n_threads=n_threads)
        p = X.cols() // K
        if groups is None:
            groups = "grouped"
        if groups == "grouped":
            groups = K * np.arange(p, dtype=int)
        elif groups == "ungrouped":
            groups = np.arange(K * p, dtype=int)
        else:
            raise RuntimeError(
                "groups must be one of \"grouped\" or \"ungrouped\" for multi-response."
            )
        group_sizes = np.concatenate([groups, [p*K]], dtype=int)
        group_sizes = group_sizes[1:] - group_sizes[:-1]
    else:
        if isinstance(X, np.ndarray):
            X = matrix.dense(X, method="naive", n_threads=n_threads)
        p = X.cols()
        if groups is None:
            groups = np.arange(p)
        group_sizes = np.concatenate([groups, [p]], dtype=int)
        group_sizes = group_sizes[1:] - group_sizes[:-1]

    if penalty is None:
        penalty = np.sqrt(group_sizes)

    etas = predict(
        X=X_raw,
        betas=betas,
        intercepts=intercepts,
        offsets=offsets,
        n_threads=n_threads,
    )

    # compute loss part
    objs = np.array([
        glm.loss(etas[i])
        for i in range(etas.shape[0])
    ])

    # relative to saturated model
    if relative:
        objs -= glm.loss_full()

    # compute regularization part
    if add_penalty:
        penalty_f = None
        if isinstance(betas, np.ndarray):
            penalty_f = core.solver.compute_penalty_dense
        elif isinstance(betas, csr_matrix): 
            penalty_f = core.solver.compute_penalty_sparse
        objs += lmdas * penalty_f(
            groups,
            group_sizes,
            penalty,
            alpha,
            betas,
            n_threads,
        )

    return objs


def residuals(
    glm: Union[GlmBase32, GlmBase64, GlmMultiBase32, GlmMultiBase64],
    etas: np.ndarray,
):
    """Computes the residuals.

    The residual is given by 
    
    .. math::
        \\begin{align*}
            \\hat{r} = -\\nabla \\ell(\\eta)
        \\end{align*}

    Parameters
    ----------
    glm : Union[glm.GlmBase32, glm.GlmBase64, glm.GlmMultiBase32, glm.GlmMultiBase64] 
        GLM object.
        It is typically one of the GLM classes defined in ``adelie.glm`` submodule.
    etas : (L, n) or (L, n, K) np.ndarray
        Linear predictions.

    Returns
    -------
    resids : (L, n) or (L, n, K) np.ndarray
        Residuals.

    See Also
    --------
    adelie.diagnostic.predict
    """
    resids = np.empty(etas.shape)
    for eta, resid in zip(etas, resids):
        glm.gradient(eta, resid)
    return resids


def gradients(
    X: Union[MatrixNaiveBase32, MatrixNaiveBase64],
    resids: np.ndarray,
    *, 
    n_threads: int =1,
):
    """Computes the gradients.

    The gradient for the single-response is given by

    .. math::
        \\begin{align*}
            \\hat{\\gamma} = X^{\\top} \\hat{r}
        \\end{align*}

    The gradient for the multi-response is given by

    .. math::
        \\begin{align*}
            \\hat{\\gamma} = (X\\otimes I_K)^{\\top} \\mathrm{vec}(\\hat{r}^\\top)
        \\end{align*}

    In both cases, :math:`\\hat{r}` is the residual as in
    ``adelie.diagnostic.residuals``.

    Parameters
    ----------
    X : (n, p) Union[MatrixNaiveBase32, MatrixNaiveBase64, np.ndarray]
        Feature matrix.
        It is typically one of the matrices defined in ``adelie.matrix`` submodule.
    resids : (L, n) or (L, n, K) np.ndarray
        Residuals.
    n_threads : int, optional
        Number of threads.
        Default is ``1``.

    Returns
    -------
    grads : (L, p) or (L, p, K) np.ndarray
        Gradients.

    See Also
    --------
    adelie.diagnostic.residuals
    """
    is_multi = len(resids.shape) == 3

    if is_multi:
        K = resids.shape[2]
        X = matrix.kronecker_eye(X, K, n_threads=n_threads)
        grad_shape = (X.cols() // K, K)
    else:
        if isinstance(X, np.ndarray):
            X = matrix.dense(X, method="naive", n_threads=n_threads)
        grad_shape = (X.cols(),)

    grads = np.empty((resids.shape[0],) + grad_shape)
    ones = np.ones(np.prod(resids.shape[1:]))
    for i in range(grads.shape[0]):
        X.mul(resids[i].ravel(), ones, grads[i].ravel())
    return grads


def gradient_norms(
    grads: np.ndarray,
    betas: csr_matrix,
    lmdas: np.ndarray,
    *, 
    groups: np.ndarray =None,
    alpha: float =1,
    penalty: np.ndarray =None,
):
    """Computes the group-wise gradient norms.

    The group-wise gradient norm is given by :math:`\\hat{h} \\in \\mathbb{R}^{G}` where

    .. math::
        \\begin{align*}
            \\hat{h}_g = \\|
                \\hat{\\gamma}_g - 
                \\lambda (1-\\alpha) \\omega_g \\beta_g
            \\|_2  \\quad g=1,\\ldots, G
        \\end{align*}

    where
    :math:`\\hat{\\gamma}_g` is the gradient as in ``adelie.diagnostic.gradients``,
    :math:`\\omega_g` is the penalty factor,
    :math:`\\lambda` is the regularization,
    :math:`\\alpha` is the elastic net proportion,
    and :math:`\\beta_g` is the coefficient block for group :math:`g`.

    .. warning::
        The group-wise gradient norm is primarily used to check the KKT conditions.
        We *do not* correct for the case of non-trivial constraints 
        since it may be too time consuming.

    Parameters
    ----------
    grads : (L, p) or (L, p, K) np.ndarray
        Gradients.
    betas : (L, p) or (L, p*K) scipy.sparse.csr_matrix
        Coefficient vectors :math:`\\beta`.
    lmdas : (L,) np.ndarray
        Regularization parameters :math:`\\lambda`.
    groups : (G,) np.ndarray, optional
        List of starting indices to each group where `G` is the number of groups.
        ``groups[i]`` is the starting index of the ``i`` th group. 
        If the gradient is of multi-response type, then we only allow two types of groupings:

            - ``"grouped"``: coefficients for each predictor is grouped across the classes.
            - ``"ungrouped"``: every coefficient is its own group.

        Default is ``None``, in which case it is set to
        ``np.arange(p)`` if ``y`` is single-response
        and ``"grouped"`` if multi-response.
    alpha : float, optional
        Elastic net parameter :math:`\\alpha`.
        It must be in the range :math:`[0,1]`.
        Default is ``1``.
    penalty : (G,) np.ndarray, optional
        Penalty factor for each group in the same order as ``groups``.
        It must be a non-negative vector.
        Default is ``None``, in which case, it is set to ``np.sqrt(group_sizes)``.

    Returns
    -------
    norms : (L, G) np.ndarray
        Gradient norms.

    See Also
    --------
    adelie.diagnostic.gradients
    """
    is_multi = len(grads.shape) == 3

    if is_multi:
        p, K = grads.shape[1:]
        if groups is None:
            groups = "grouped"
        if groups == "grouped":
            groups = K * np.arange(p, dtype=int)
        elif groups == "ungrouped":
            groups = np.arange(K * p, dtype=int)
        else:
            raise RuntimeError(
                "groups must be one of \"grouped\" or \"ungrouped\" for multi-response."
            )
        group_sizes = np.concatenate([groups, [p*K]], dtype=int)
        group_sizes = group_sizes[1:] - group_sizes[:-1]
    else:
        p = grads.shape[-1]
        if groups is None:
            groups = np.arange(p)
        group_sizes = np.concatenate([groups, [p]], dtype=int)
        group_sizes = group_sizes[1:] - group_sizes[:-1]

    if penalty is None:
        penalty = np.sqrt(group_sizes)
    penalty = np.repeat(penalty, group_sizes)

    L = grads.shape[0]
    grads = grads.reshape((L, -1)) - betas.multiply(lmdas[:, None] * (1 - alpha) * penalty[None])
    return np.array([
        np.linalg.norm(grads[:, g:g+gs], axis=-1)
        for g, gs in zip(groups, group_sizes)
    ]).T


def gradient_scores(
    grad_norms: np.ndarray,
    lmdas: np.ndarray,
    *, 
    alpha: float =1,
    penalty: np.ndarray =None,
):
    """Computes the gradient scores.

    The gradient score is given by

    .. math::
        \\begin{align*}
            \\hat{s}_g = 
            \\begin{cases}
                \\hat{h}_g \\cdot (\\alpha p_g)^{-1} ,& \\alpha p_g > 0 \\\\
                \\lambda ,& \\alpha p_g = 0
            \\end{cases}
            \\qquad
            g = 1,\\ldots, G
        \\end{align*}

    where :math:`\\hat{h}` is the gradient norm as in
    ``adelie.diagnostic.gradient_norms``.

    Parameters
    ----------
    grad_norms : (L, G) np.ndarray
        Gradient norms.
    lmdas : (L,) np.ndarray
        Regularization parameters :math:`\\lambda`.
    penalty : (G,) np.ndarray 
        Penalty factor for each group.
        It must be a non-negative vector.
    alpha : float, optional
        Elastic net parameter :math:`\\alpha`.
        It must be in the range :math:`[0,1]`.
        Default is ``1``.

    Returns
    -------
    scores : (L, G) np.ndarray
        Gradient scores.  

    See Also
    --------
    adelie.diagnostic.gradient_norms
    """
    denom = alpha * penalty
    scores = np.divide(grad_norms, denom[None], where=denom[None] > 0)
    scores[:, denom <= 0] = lmdas[:, None]
    return scores


def coefficient(
    lmda: float,
    betas: csr_matrix,
    intercepts: np.ndarray,
    lmdas: np.ndarray,
):
    """Computes the coefficient at :math:`\\lambda` using linear interpolation of solutions.

    The linearly interpolated coefficient is given by
    
    .. math::
        \\begin{align*}
            \\hat{\\beta}(\\lambda)
            =
            \\frac{\\lambda - \\lambda_{k+1}}{\\lambda_{k} - \\lambda_{k+1}}
            \\hat{\\beta}(\\lambda_k)
            +
            \\frac{\\lambda_{k} - \\lambda}{\\lambda_{k} - \\lambda_{k+1}}
            \\hat{\\beta}(\\lambda_{k+1})
        \\end{align*}

    if :math:`\\lambda \\in [\\lambda_{k+1}, \\lambda_k]`.
    If :math:`\\lambda` lies above the largest value in ``lmdas`` or below the smallest value,
    then we simply take the solution at the respective ends.
    The same formula holds for intercepts.

    Parameters
    ----------
    lmda : float
        New regularization parameter at which to find the solution.
    betas : (L, p) scipy.sparse.csr_matrix
        Coefficient vectors :math:`\\beta`.
    intercepts : (L,) np.ndarray
        Intercepts.
    lmdas : (L,) np.ndarray
        Regularization parameters :math:`\\lambda`.

    Returns
    -------
    beta : (1, p) scipy.sparse.csr_matrix
        Linearly interpolated coefficient vector at :math:`\\lambda`.
    intercept : float
        Linearly interpolated intercept at :math:`\\lambda`.
    """
    if len(lmdas) == 0:
        raise RuntimeError("lmdas must be non-empty!")
    if len(lmdas) == 1:
        return betas, lmdas
    order = np.argsort(lmdas)
    idx = np.searchsorted(
        lmdas,
        lmda,
        sorter=order,
    )
    idx = lmdas.shape[0] - idx
    if idx == 0 or idx == lmdas.shape[0]:
        logger.logger.warning(
            "lmda is not within the range of the saved lambdas. " +
            "Returning boundary solution."
        )
        idx = np.clip(idx, 0, lmdas.shape[0]-1)
        return betas[idx], intercepts[idx]

    left, right = betas[idx-1], betas[idx]
    weight = (lmda - lmdas[idx]) / (lmdas[idx-1] - lmdas[idx])
    beta = left.multiply(weight) + right.multiply(1-weight)
    left, right = intercepts[idx-1], intercepts[idx]
    intercept = weight * left + (1-weight) * right

    return beta, intercept


def plot_coefficients(
    betas: csr_matrix,
    lmdas: np.ndarray,
    groups: np.ndarray,
    group_sizes: np.ndarray,
    *,
    l2_norm: bool =False,
):
    """Plots the coefficient profile.

    Parameters
    ----------
    betas : (L, p) scipy.sparse.csr_matrix
        Coefficient vectors :math:`\\beta`.
    lmdas : (L,) np.ndarray
        Regularization parameters :math:`\\lambda`.
    groups : (G,) np.ndarray
        List of starting indices to each group where `G` is the number of groups.
        ``groups[i]`` is the starting index of the ``i`` th group. 
    group_sizes : (G,) np.ndarray
        List of group sizes corresponding to each element of ``groups``.
        ``group_sizes[i]`` is the size of the ``i`` th group.
    l2_norm : bool, optional
        If ``True``, the :math:`\\ell_2` norms of each coefficient group
        is plotted rather than the raw coefficient values.
        This may be more intuitive to visualize since there is only one path
        associated with a group.
        Default is ``False``.

    Returns
    -------
    fig, ax
    """
    tls = -np.log(lmdas)

    fig, ax = plt.subplots(figsize=(9, 6), layout="constrained")
    colors = plt.rcParams["axes.prop_cycle"].by_key()["color"]
    colors_it = cycle(colors)

    if l2_norm:
        for g, gs in zip(groups, group_sizes):
            curr_block = betas[:, g:g+gs]
            if curr_block.nnz == 0:
                continue
            curr_block = curr_block.toarray()
            curr_block_norms = np.linalg.norm(curr_block, axis=-1)
            color = next(colors_it)
            ax.plot(tls, curr_block_norms, linestyle="-", color=color)

        ax.set_title(r"Coefficient $\ell_2$-Norm Profile")
        ax.set_ylabel(r"$\|\beta\|_2$")
        ax.set_xlabel(r"-$\log(\lambda)$")
    else:
        for g, gs in zip(groups, group_sizes):
            curr_block = betas[:, g:g+gs]
            if curr_block.nnz == 0:
                continue
            curr_block = curr_block.toarray()
            color = next(colors_it)
            ax.plot(tls, curr_block, linestyle="-", color=color)

        ax.set_title(r"Coefficient Profile")
        ax.set_ylabel(r"$\beta$")
        ax.set_xlabel(r"-$\log(\lambda)$")

    return fig, ax


def plot_devs(
    lmdas: np.ndarray,
    devs: np.ndarray,
):
    """Plots the deviance profile.

    Parameters
    ----------
    lmdas : (L,) np.ndarray
        Regularization parameters :math:`\\lambda`.
    devs : (L,) np.ndarray
        Deviances.

    Returns
    -------
    fig, ax
    """
    tls = -np.log(lmdas)

    fig, ax = plt.subplots(figsize=(9, 6), layout="constrained")
    ax.plot(tls, devs, linestyle='-', color='r', marker='.')
    ax.set_title(r"Deviance Profile")
    ax.set_ylabel(r"Deviance Explained (\%)")
    ax.set_xlabel(r"$-\log(\lambda)$")

    return fig, ax


def plot_set_sizes(
    groups: np.ndarray,
    screen_sizes: np.ndarray,
    active_sizes: np.ndarray,
    lmdas: np.ndarray,
    screen_rule: str,
    *,
    ratio: bool =False,
    exclude: list =[],
    axes = None,
):
    """Plots the set sizes.

    Parameters
    ----------
    groups : (G,) np.ndarray
        List of starting indices to each group where `G` is the number of groups.
        ``groups[i]`` is the starting index of the ``i`` th group. 
    screen_sizes : (L,) np.ndarray
        Screen set sizes.
    active_sizes : (L,) np.ndarray
        Active set sizes.
    lmdas : (L,) np.ndarray
        Regularization parameters :math:`\\lambda`.
    ratio : bool, optional
        ``True`` if plot should normalize the set sizes
        by the total number of groups.
        Default is ``False``.
    exclude : list, optional
        The sets to exclude from plotting.
        It must be a subset of the following:

            - ``"active"``: active set.
            - ``"screen"``: screen set.

        Default is ``[]``.
    axes
        Matplotlib axes object.

    Returns
    -------
    fig, ax
        If ``axes`` is ``None``, both are returned.
    ax
        If ``axes`` is not ``None``, then only ``ax`` is returned.
    """
    make_ax = axes is None
    ax = axes

    include = ["active", "screen"]
    if len(exclude) > 0:
        include = list(set(include) - set(exclude))
    
    include_map = {
        "active": 0,
        "screen": 1,
    } 

    ys = [
        active_sizes,
        screen_sizes,
    ]
    if ratio:
        ys = [y / len(groups) for y in ys]

    labels = [
        "active",
        screen_rule,
    ]
    colors = [
        "tab:red",
        "tab:blue",
    ]
    markers = ["o", "v"]

    y_sizes = np.array([y.shape[0] for y in ys])
    iters = np.min(y_sizes)
    if not np.all(y_sizes == iters):
        logger.logger.warning(
            "The sets do not all have the same set sizes. " +
            "The plot will only show up to the smallest set."
        )
    tls = -np.log(lmdas[:iters])
    ys = [y[:iters] for y in ys]

    ys = [ys[include_map[s]] for s in include]
    labels = [labels[include_map[s]] for s in include]
    colors = [colors[include_map[s]] for s in include]
    markers = [markers[include_map[s]] for s in include]

    if make_ax:
        fig, ax = plt.subplots(figsize=(9, 6), layout="constrained")

    for y, marker, color, label in zip(ys, markers, colors, labels):
        ax.plot(
            tls,
            y, 
            linestyle="None", 
            marker=marker,
            markerfacecolor="None",
            color=color,
            label=label,
        )
    ax.legend()
    ax.set_title("Set Size Profile")
    if ratio:
        ax.set_ylabel("Proportion of Groups")
    else:
        ax.set_ylabel("Number of Groups")
    ax.set_xlabel(r"$-\log(\lambda)$")

    if make_ax:
        return fig, ax
    return ax


def plot_benchmark(
    total_time: np.ndarray,
    benchmark_screen: np.ndarray,
    benchmark_fit_screen: np.ndarray,
    benchmark_fit_active: np.ndarray,
    benchmark_kkt: np.ndarray,
    benchmark_invariance: np.ndarray,
    n_valid_solutions: np.ndarray,
    lmdas: np.ndarray,
    *,
    relative: bool =False,
):
    """Plots benchmark times.

    Parameters
    ----------
    total_time : float
        Total time taken for the core routine.
    benchmark_screen : (B,) np.ndarray
        Benchmark timings for screening.
    benchmark_fit_screen : (B,) np.ndarray
        Benchmark timings for fitting on screen set.
    benchmark_fit_active : (B,) np.ndarray
        Benchmark timings for fitting on active set.
    benchmark_kkt : (B,) np.ndarray
        Benchmark timings for KKT checks.
    benchmark_invariance : (B,) np.ndarray
        Benchmark timings for invariance step.
    n_valid_solutions : (B,) np.ndarray
        Flags that indicate whether each iteration resulted in a valid solution.
    lmdas : (L,) np.ndarray
        Regularization parameters :math:`\\lambda`.
    relative : bool, optional
        If ``True``, the time breakdown plot is relative to the total time,
        therefore plotting the proportion of time spent in each category.
        Otherwise, the absolute times are shown.
        Default is ``False``.

    Returns
    -------
    fig, ax
    """
    def _squash_times(ts):
        idx = 0
        new_ts = []
        while idx < len(n_valid_solutions):
            if n_valid_solutions[idx]:
                new_ts.append(ts[idx]) 
            else:
                t = 0
                while (idx < len(n_valid_solutions)) and (not n_valid_solutions[idx]):
                    t += ts[idx]
                    idx += 1
                if idx < len(n_valid_solutions):
                    t += ts[idx]
                new_ts.append(t)
            idx += 1
        return np.array(new_ts)

    times = [
        _squash_times(benchmark_screen),
        _squash_times(benchmark_fit_screen),
        _squash_times(benchmark_fit_active),
        _squash_times(benchmark_kkt),
        _squash_times(benchmark_invariance),
    ]
    n_iters = np.min([lmdas.shape[0]] + [len(t) for t in times])
    times = [t[:n_iters] for t in times]
    lmdas = lmdas[:n_iters]
    tlmdas = -np.log(lmdas)

    colors = [
        "green",
        "orange",
        "red",
        "purple",
        "brown",
    ]
    markers = [
        ".", "v", "^", "+", "*",
    ]
    labels = [
        "screen", "fit-screen", "fit-active", "kkt", "invariance",
    ]

    fig, axes = plt.subplots(1, 2, figsize=(9, 6), layout="constrained")
    for tm, color, marker, label in zip(times, colors, markers, labels):
        axes[0].plot(
            tlmdas,
            tm,
            linestyle="None",
            color=color,
            marker=marker,
            markerfacecolor="None",
            label=label,
        )
    axes[0].legend()
    axes[0].set_title("Benchmark Profile")
    axes[0].set_ylabel("Time (s)")
    axes[0].set_xlabel(r"$-\log(\lambda)$")
    axes[0].set_yscale("log")

    total_times = np.array([np.sum(t) for t in times])
    total_times_sum = np.sum(total_times)
    total_times = np.concatenate([
        total_times, 
        [total_time - total_times_sum] # unaccounted time
    ])
    if relative:
        total_times /= total_time
    axes[1].bar(
        np.arange(len(total_times)),
        total_times,
        color=colors + ["grey"],
        edgecolor=colors + ["grey"],
        linewidth=1.5,
        label=labels + ["other"],
        alpha=0.5,
    )
    axes[1].legend()
    axes[1].set_title("Time Breakdown")
    if relative:
        axes[1].set_ylabel("Proportion of Time")
    else:
        axes[1].set_ylabel("Time (s)")
    axes[1].set_xlabel("Category")

    return fig, axes


def plot_kkt(
    lmdas: np.ndarray,
    scores: np.ndarray, 
    *,
    idx: int =None,
    relative: bool =False,
):
    """Plots KKT failures.

    This function only plots a subset of the scores.
    Specifically (assuming ``idx`` is not ``None``),
    if there are ``n`` number of scores in ``scores[idx]``
    above the threshold ``lmdas[idx]``, then these ``n`` scores 
    as well as the ``n`` largest scores below the threshold are shown.
    If ``idx`` is ``None``, then the same plot is drawn at every index
    as an animation.

    Parameters
    ----------
    lmdas : (L,) np.ndarray
        Regularization parameters :math:`\\lambda`.
    scores : (L, G) np.ndarray
        Gradient scores.
    idx : int, optional
        Index of ``lmdas`` and ``scores`` at which to plot the KKT failures.
        If ``None``, then an animation of the plots at every index is shown.
        Default is ``None``.
    relative : bool, optional
        If ``True``, then plots the relative error ``score / lmda[:, None] - 1``.
        Otherwise, the absolute values are used.
        Default is ``False``.

    Returns
    -------
    fig, ax
        If ``idx`` is not ``None``, then a figure and axes is returned.
    anim
        If ``idx`` is ``None``, then an animation of the plots is returned.
    """
    G = scores.shape[-1]

    if relative:
        scores = scores / lmdas[:, None] - 1
        baseline = np.zeros(scores.shape[0])
    else:
        baseline = lmdas

    do_anim = idx is None
    idx = 0 if do_anim else idx

    gns = np.arange(G)

    colors = ["blue", "red"]
    labels = ["success", "failure"]
    alphas = [0.6, 0.8]

    fig, ax = plt.subplots(figsize=(9, 6), layout="constrained")

    is_failure = scores[idx] > baseline[idx]
    xs = [
        gns[~is_failure],
        gns[is_failure],
    ]
    ys = [
        scores[idx, ~is_failure],
        scores[idx, is_failure],
    ]
    scats = [None] * 2
    for i, (x, y, color, label, alpha) in enumerate(zip(xs, ys, colors, labels, alphas)):
        scats[i] = ax.scatter(
            x, y,
            color=color,
            marker='.',
            facecolor="None",
            label=label,
            alpha=alpha,
        )
    ax.legend()
    bound = np.maximum((np.max(scores[idx]) - baseline[idx]) * 1.05, 1e-5)
    ax.set_ylim(
        bottom=baseline[idx]-bound,
        top=baseline[idx]+bound,
    )
    ax.axhline(baseline[idx], linestyle='--', linewidth=1, color="green")
    ax.set_title("Active Score Error (Largest)")
    if relative:
        ax.set_ylabel(r"$s_g / \lambda - 1")
    else:
        ax.set_ylabel(r"$s_g$")
    ax.set_xlabel("Group Number")

    if do_anim:
        plt.close(fig)
    else:
        return fig, ax

    def update(idx):
        s = scores[idx]

        is_failure = s > baseline[idx]
        xs = [
            gns[~is_failure],
            gns[is_failure],
        ]
        ys = [
            s[~is_failure],
            s[is_failure],
        ]
        for i, (x, y, color, label, alpha) in enumerate(zip(xs, ys, colors, labels, alphas)):
            data = np.stack([x, y]).T
            scats[i].set_offsets(data)
            scats[i].set(
                color=color,
                facecolor="None",
                alpha=alpha,
                label=label,
            )
        bound = np.maximum((np.max(s) - baseline[idx]) * 1.05, 1e-5)
        ax.set_ylim(
            bottom=baseline[idx]-bound,
            top=baseline[idx]+bound,
        )
        return (scats[0], scats[1],)

    anim = animation.FuncAnimation(
        fig=fig, 
        func=update, 
        frames=lmdas.shape[0]-1, 
        interval=200, 
        repeat=False,
    )

    return HTML(anim.to_html5_video())


class DiagnosticCov:
    """Diagnostic class for covariance states.

    .. note::
        Currently, the only supported covariance state
        is the Gaussian covariance state from calling ``adelie.gaussian_cov``.

    Parameters
    ----------
    state
        A state object of covariance type from solving group elastic net.
    """
    def __init__(self, state):
        self.state = state
        self.betas = state.betas

        # keep same format as DiagnosticNaive for consistency
        self._args = {}
        self._args["groups"] = state.groups
        self._args["penalty"] = state.penalty

        A, v = state.A, state.v
        L, p = self.betas.shape
        self.gradients = np.empty((L, p))
        for i in range(L):
            beta_i = self.betas[i]
            A.mul(beta_i.indices, beta_i.data, self.gradients[i]) 
        self.gradients = v[None] - self.gradients
        self.gradient_norms = gradient_norms(
            grads=self.gradients,
            betas=self.betas,
            lmdas=self.state.lmdas,
            groups=self._args["groups"],
            alpha=self.state.alpha,
            penalty=self._args["penalty"],
        )
        self.gradient_scores = gradient_scores(
            grad_norms=self.gradient_norms,
            lmdas=self.state.lmdas,
            alpha=self.state.alpha,
            penalty=self._args["penalty"],
        )

    def plot_coefficients(self, **kwargs):
        """Plots the coefficient profile.

        See Also
        --------
        adelie.diagnostic.plot_coefficients
        """
        return plot_coefficients(
            betas=self.betas,
            lmdas=self.state.lmdas,
            groups=self.state.groups,
            group_sizes=self.state.group_sizes,
            **kwargs,
        )

    def plot_devs(self):
        """Plots the deviance profile.

        See Also
        --------
        adelie.diagnostic.plot_devs
        """
        return plot_devs(
            lmdas=self.state.lmdas,
            devs=self.state.devs,
        )

    def plot_set_sizes(self, **kwargs):
        """Plots the set sizes.

        See Also
        --------
        adelie.diagnostic.plot_set_sizes
        """
        return plot_set_sizes(
            groups=self.state.groups,
            screen_sizes=self.state.screen_sizes,
            active_sizes=self.state.active_sizes,
            lmdas=self.state.lmdas,
            screen_rule=self.state.screen_rule,
            **kwargs,
        )

    def plot_benchmark(self, **kwargs):
        """Plots benchmark times.

        See Also
        --------
        adelie.diagnostic.plot_benchmark
        """
        return plot_benchmark(
            total_time=self.state.total_time,
            benchmark_screen=self.state.benchmark_screen,
            benchmark_fit_screen=self.state.benchmark_fit_screen,
            benchmark_fit_active=self.state.benchmark_fit_active,
            benchmark_kkt=self.state.benchmark_kkt,
            benchmark_invariance=self.state.benchmark_invariance,
            n_valid_solutions=self.state.n_valid_solutions,
            lmdas=self.state.lmdas,
            **kwargs,
        )

    def plot_kkt(self, **kwargs):
        """Plots KKT failures.

        See Also
        --------
        adelie.diagnostic.plot_kkt
        """
        return plot_kkt(
            lmdas=self.state.lmdas,
            scores=self.gradient_scores,
            **kwargs,
        )
    

class DiagnosticNaive:
    """Diagnostic class for naive states.

    Parameters
    ----------
    state
        A state object of naive type from solving group elastic net.
    """
    def __init__(self, state):
        self.state = state
        self.betas = state.betas
        self._n_classes = self.state._glm.y.shape[-1]
        self._is_multi = state._glm.is_multi
        self._args = {}
        if self._is_multi:
            self._args["groups"] = state.group_type
            p_begin = self.state.multi_intercept * self._n_classes
            self._args["penalty"] = state.penalty[p_begin:]
        else:
            self._args["groups"] = state.groups
            self._args["penalty"] = state.penalty

        self.linear_preds = predict(
            X=self.state._X,
            betas=self.betas,
            intercepts=self.state.intercepts,
            offsets=self.state._offsets,
            n_threads=self.state.n_threads,
        )
        self.residuals = residuals(
            glm=self.state._glm,
            etas=self.linear_preds,
        )
        self.gradients = gradients(
            X=self.state._X,
            resids=self.residuals,
            n_threads=self.state.n_threads,
        )
        self.gradient_norms = gradient_norms(
            grads=self.gradients,
            betas=self.betas,
            lmdas=self.state.lmdas,
            groups=self._args["groups"],
            alpha=self.state.alpha,
            penalty=self._args["penalty"],
        )
        self.gradient_scores = gradient_scores(
            grad_norms=self.gradient_norms,
            lmdas=self.state.lmdas,
            alpha=self.state.alpha,
            penalty=self._args["penalty"],
        )

    def plot_coefficients(self, **kwargs):
        """Plots the coefficient profile.

        See Also
        --------
        adelie.diagnostic.plot_coefficients
        """
        p_begin = (
            self.state.multi_intercept * self._n_classes
            if self._is_multi else
            0
        )
        return plot_coefficients(
            betas=self.betas,
            lmdas=self.state.lmdas,
            groups=self.state.groups[p_begin:]-p_begin,
            group_sizes=self.state.group_sizes[p_begin:],
            **kwargs,
        )

    def plot_devs(self):
        """Plots the deviance profile.

        See Also
        --------
        adelie.diagnostic.plot_devs
        """
        return plot_devs(
            lmdas=self.state.lmdas,
            devs=self.state.devs,
        )

    def plot_set_sizes(self, **kwargs):
        """Plots the set sizes.

        See Also
        --------
        adelie.diagnostic.plot_set_sizes
        """
        return plot_set_sizes(
            groups=self.state.groups,
            screen_sizes=self.state.screen_sizes,
            active_sizes=self.state.active_sizes,
            lmdas=self.state.lmdas,
            screen_rule=self.state.screen_rule,
            **kwargs,
        )

    def plot_benchmark(self, **kwargs):
        """Plots benchmark times.

        See Also
        --------
        adelie.diagnostic.plot_benchmark
        """
        return plot_benchmark(
            total_time=self.state.total_time,
            benchmark_screen=self.state.benchmark_screen,
            benchmark_fit_screen=self.state.benchmark_fit_screen,
            benchmark_fit_active=self.state.benchmark_fit_active,
            benchmark_kkt=self.state.benchmark_kkt,
            benchmark_invariance=self.state.benchmark_invariance,
            n_valid_solutions=self.state.n_valid_solutions,
            lmdas=self.state.lmdas,
            **kwargs,
        )

    def plot_kkt(self, **kwargs):
        """Plots KKT failures.

        See Also
        --------
        adelie.diagnostic.plot_kkt
        """
        return plot_kkt(
            lmdas=self.state.lmdas,
            scores=self.gradient_scores,
            **kwargs,
        )


def diagnostic(state):
    """Creates a diagnostic class object appropriate for the state.

    Parameters
    ----------
    state
        A state object from solving group elastic net.
    
    Returns
    -------
    dg
        Diagnostic class object.

    See Also
    --------
    adelie.diagnostic.DiagnosticCov
    adelie.diagnostic.DiagnosticNaive
    """
    if "naive" in type(state).__name__:
        return DiagnosticNaive(state)
    elif "cov" in type(state).__name__:
        return DiagnosticCov(state)
    raise TypeError("state must be one of the supported state types in adelie.")
